//version: 4.7
using System;
using System.Collections;
using System.Text;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.Globalization;
using System.Reflection;
using Dbasic;
using System.Collections.Generic;
using SysThread = System.Threading;

namespace BasicLib
{
	public class BasicLib
	{
		// private variables

		// 1.1 amended RunArray to avoid <private implementation> showing in B4ppc
		//     implemented Select ... Case ..End Select
		//     slightly improved tokeniser keyword/sub loop
		//     added underscore token to allow use in Sub names
		//     fixed SubString, StrAt, allow uppercase variables
		// 1.2 pre-tokenised program, new getnexttoken and other necessary mods.
		//     about 16 x faster than 1.0 and 8 X faster than 1.1
		//     now allows Return within If, For or Do loop
		//     now allows Else If
		//     CTicksPerSecond fixed
		//     changed treatment of Args to set individual variables
		//     Starting event, GetArray and SetArray methods added
		// 2.0 user defined variables implemented
		//     tokenised variable index into source so no lookups at run time
		// 2.1 CallHost errors still referenced Msgbox. Now fixed.
		//     fixed Rem broken by pre-tokenising
		//     fixed Date functions and Time functions broken by new variable code
		//     added ? as synonym for Print
		//     added Goto and labels
		//     various bugs and ommisions and performance fixes
		// 2.2 Sleep & Version added. all numeric Parse and ToString now use the "cul" variant
		// 2.3 Precedence of the power operator increased over * / MOD
		//     Round didn't have optional parameter, unary- was a string not numeric operation
		//     unary+ now treated as "0+", unary- now treated as "0-"
		//     fixed IsPunctuation name (was IsPunc), Format didn't call getnexttoken on entry
		//     tokeniser now accepts scientific notation
		//     Dim now checks for variable being global
		// 2.4 A stupid typo broke MOD in version 2.3
		//     removed redundant callstack struct and changed callstack to type int 
		//     added Halt method to stop an endless loop
		//     Rem, EndSub (not "End Sub") had problems, now fixed
		//     error now raised on attempted array acess to a local variable
		// 3.0 now allows Sub calls in called Sub parameter lists
		//     added Tokens property to access tokenised source code  
		//     new higher performance line token structure
		//     second pass added to precalculate jump destinations for IF, SELECT, FOR and DO
		// 3.1 CallHostSub statement and function added
		// 3.2 Excess CallHostSubx Enum elements removed
		//     CallHostGuiSub statement and function added
		//     CallSub statement and function added
		//     RGB, GetR, GetG, GetB and the Color constants  added
		//     Step and Break now stop before a line is executed as in the IDE and Runner
		// 3.3 added ShareArray and Rnd - ommitted in error
		// 3.4 CallHostSub and CallHostGuiSub now work in the IDE also previously were the wrong way round
		// 3.5 fixed tokeniservariable check for local array which was broken
		//     fixed Callsub error when used to call a Sub with no parameters
		//     fixed ArrayCopy bug source and target starts interchanged
		//     improved performance by looking up string literals and Sub names from an ArrayList instead of inline
		//     added Getitem, SetItem, GetRank and GetTotalLen indirect array access in lieu of passing arrays as parameters
		//     added decXXX decompiler helper methods and properties
		//     added Dim Type(v1, ...) and . operator for array access to simulate B4P types
		//     added multi-dimensional arrays, ArrayRank and ArrayTotalLen statement
		// 4.0 added support for new 6.90 variable and Sub structures
		//     removed need for B4PObject(1) in constructor
		// 4.1 for some reason LoadCodeFrom Array called initvars - no more!
		//     now initialises globals to an empty string as the documentation says!
		//     calling a Sub in a Do While expression failed by altering linenum - now fixed
		// 4.2 now matches closing string character delimiter to the opening one 
		// 4.3 Goto broke a few versions ago.
		//     Array always set the array length to 1 instead of the actual length
		//     corrected SetArray and ShareArray to set correct dimension size
		// 4.4 corrected StrSplit to set correct array dimension size
		//		added StrJoin(Array, Delimiter)
		//      default dateformat now "dd/MM/yyyy";
		//      added # as synonym for REM
		//      initvars() was not setting the length of the args array.
		//      added LoadCodeAsString2, SourceLineNumber and LoadedLineNumber
		//      added Call and CanCall from Android BasicLibIDE
		//      corrected typo of StrRemove
		//      added missing return after gotolinenum(...) in elseifstatement()
		//      variables can now start with underscore
		//		array access DOT syntax now accepts numeric literals and multi-dimensioned arrays
		// 4.5  added AddSubCalls, GetSubCalls & ClearSubCalls to Basic4ppc interface to allow hosts Subs to be called directly
		// 4.6  fixed casts in the first two lines of gotostatement - were (int) which resulted in wrong +ves like 655535 instead of -1
		//      save ifstackptr in For and Do, restored in Exit for when Exit is called within one or more nested If blocks
		// 4.7  changed development methods and properties prefix from dec to dev
		//      added devSubNames to return a string array of the Sub names
		//      added devGlobals to return a string array of the Global names
		//      added devLocals to return a string array of the Local names for a given Sub
		private double version = 4.7;

		// code settings
		private Random random;
		private String dateformat;
		private String timeformat;
		private static CultureInfo cul = new CultureInfo("en-US", false);

	   
		// debug and messaging
		private bool stepping = false;
		private int breakline;
		private String messagestring = "";
		private String subjectstring = "";
		private String errorstring = "";

		// program source code
		private String source; // program
		private String[] sourcearray; // program

		// program execution
		private int tokenptr; // the program counter, points to token start
		private int nextptr; // offset to end of current token + 1
		private String subname;// when tokeniser recognizes a SUB name 
		private String returnval;// set by Return
		private _Token currenttoken; // the present token
		private int linenum; // present line number
		private bool ended;

		// variables and parameters
		private String[][] globals; // [Var][ArrayIndex]
		private int[][] globalsrank; // [Var][Dimension]
		private String[,] locals; // [CallStackPtr-1, Var]
		private ArrayList[] parameters; // [CallStackPtr-1]


		// program data structures
		private Hashtable subptrs;  // key - string Sub name : value int linenumber
		private Hashtable lineptrs; // key - int linenumber : value int position of line number
		private ArrayList statements; // the statement token of each line in linenumber order   
		private Hashtable labelptrs; // key - string label name : value int linenum
		private ArrayList literals; // literal strings, numbers and Sub names used by the program   
 
		private Hashtable globalvars; // key - string global name : value int index in globals
		private Hashtable subvars; // key - string Sub name : value Hashtable localvars
		private Hashtable localvars; // key - string local name : value int index in locals
		private int maxlocals; // maximum number of locals in any sub

		// keywords and tokens
		private Hashtable keywords;  // key - string keyword : value - Token token
		private Hashtable hostsubnames; // key -  subname : value - Token token


		private enum _Token {
			ENDOFINPUT, ERROR, CR, NULL, REM, APPCLOSE, CALLHOST, SLEEP, //0-7
			LET, NUMBER, STRING, VARIABLE, DIM, TYPE, // 8-13 
			ARRAY, ARRAYLEN, ARRAYCOPY, ARRAYTOTALLEN,ARRAYRANK, //14-18        
			FOR, TO, STEP, NEXT, DO, LOOP, WHILE, UNTIL, EXIT, GOTO, //19-28
			IF, THEN, ELSE, END, CALL, SUB, SELECT, ENDIF, ENDSELECT, ENDSUB, //29-38 
			ELSEIF, RETURN,  CASE, SHARP, LEFTPAREN, RIGHTPAREN, LABEL, //39-45        
			PRINT, QUERY, COMMA, DOT, SEMICOLON, MSGBOX, UNDERSCORE, //46-52
			PLUS, MINUS, BITAND, BITOR, BITXOR, AND, OR, POWER, APPEND, //53-61
			TIMES, DIVIDE, MOD, TRUE, FALSE, LT, GT, EQ, NE, LE, GE, //62-72  
			GETITEM, SETITEM, GETRANK, GETLEN, GETTOTALLEN, //73-77 
			COS, SIN, TAN, ACOS, ASIN, ATAN, SQRT, INT, NOT, //78-86
			CPI, CE, CRLF, CTAB, CASESENSITIVE, CASEUNSENSITIVE, VERSION, //87-93
			CTICKSPERDAY, CTICKSPERHOUR, CTICKSPERMINUTE, CTICKSPERSECOND, NOW, //94-98
			RND, ABS, MAX, MIN, LN, LOG, ROUND, GETR, GETG, GETB, RGB, //99-109
			DATEFORMAT, DATE, DATEADD, DATEDAY, DATEDAYOFWEEK, DATEDAYOFYEAR, //110-115
			DATEMONTH, DATEPARSE, DATEYEAR, //116-118
			TIMEFORMAT, TIME, TIMEADD, TIMEHOUR, TIMEMINUTE, TIMEPARSE, TIMESECOND, //119-125 
			ASC, CHR, ISDIGIT, ISLETTER, ISNUMBER, ISPUNC, STRAT, FORMAT, //126-133
			STRCOMPARE, STRINDEXOF, STRINSERT, STRLENGTH, STRREMOVE, //134-138
			STRREPLACE, STRTOLOWER, STRTOUPPER, SUBSTRING, STRSPLIT, //139-143
			CALLHOSTSUB, CALLHOSTGUISUB, CALLSUB, CBEIGE, CBLACK, CBLUE, CBROWN, CCYAN, //144-151
			CGOLD, CGREEN, CGRAY, CORANGE, CPINK, CPURPLE, CRED, CSILVER, CWHITE, CYELLOW, //152-161
			STRJOIN, HOSTSUBCALL//162-163
			// All global variables can be arrays of rank 1, 2 or 3
			// MsgBox takes one or two parameters and causes a Message event,
			// Print or ? formats a string and causes a Message event Subject "Print"
			// Strings can be single or double quotes delimited
			// Only multi-line If statements are allowed
			// Rem must start a statement on a new line
			// Colon statement separator not supported
			//
			// Keywords implemented
			// --------------------
			//
			// StrSplit(arrayvar, string, seps) has no return value
			// StrJoin(arrayvar, separator) return value is String
			// Array(var, val, val, ...) has no return value
			//
			// BitAnd, BitOr, BitXor operators implemented
			//
			// General : all except B4PObject, DoEvents. ErrorLabel, IsNull
			//           SetTransparent, Shell, SIP, Sound, WaitCursor
			// Math : all
			// String : all
			// Time : except DateD, DateM, DateY, TimeH, Time M, TimeS
			// Constants : except Data, File, Image, MsgBox, SpecialKey
		}

		// multi-statement preprocessing
		private int[] sourcelines; // index loaded line, value source line for LoadCodeAsStringMultiStatement
		private int[] loadedlines; // index souce line, value loaded line for LoadCodeAsStringMultiStatement
  
		// stacks
		private int MaxStackDepth = 5;

		// local variables
		private Hashtable[] localsstack; // the localvars hashtable for the call level = [SubDepth]

		// Select ... End Select
		private String[] selectvals; // the value of a Select statment = [SubDepth]
		private int[] endselectlines; // the line number of End Select = [SubDepth]
		private bool[] matchedcases; // true if a matched code executed = [SubDepth]

		// Call ... Return
		private int[] callstack; // SubReturn = callstackptr[SubDepth]
		private int callstackptr; // SubDepth
		private int localsptr; // SubDepth usually except for parameter calculation

		//For ... Next
		private _forstruct[,] forstack; // ForStruct = forstack[SubDepth, ForDepth]
		private int[] forstackptr; // ForDepth = [SubDepth]
		private struct _forstruct
		{
			public int loopline;  // line number destination for looping
			public int exitline;  // line number for exit statement
			public int[] forvariable; // the for loop variable
			public double to; // the end loop value           
			public double step; // the step by value
			public int ifstackptr; // value of ifstackptr on entry to For, restored by Exit keyword
		};

		//If ... End If
		private _ifstruct[,] ifstack; // IfStruct = ifstackptr[SubDepth, IfDepth]
		private int[] ifstackptr; // IfDepth = [SubDepth]
		private struct _ifstruct
		{
			public int endifline;  // End If line number
			public String val;  // boolean result of If test
		};

		// B4ppc events
		private object[] eventObject;
		public event EventHandler Break; //The event handler name must match the event name.
		public event EventHandler CallHost; //The event handler name must match the event name.
		public event EventHandler Loaded; //The event handler name must match the event name.
		public event EventHandler Stepped; //The event handler name must match the event name.
		public event EventHandler Starting; //The event handler name must match the event name.
		public event EventHandler Message; //The event handler name must match the event name.
		public event EventHandler Ended; //The event handler name must match the event name.

		MethodInfo miStartSub; // StartSub(Place subsPlace, Dictionary<string, RegularVariable> localsReg, Sub currentSub)
		MethodInfo miMakeRegularVariable; // MakeRegularVariable(Type type)
		MethodInfo miMakeRegularDictionary; // MakeRegularDictionary()
		private IDictionary htSubs; // holds the Dbasic delegates or Sub objects
		object b4p; // CRunner or b4p
		private bool legacy; // true if not compiled
		private string prefix; // b4ppc internal name prefix
		private delegate object _mrunnerdel(Object o1, MethodInfo o2, Object[] o3);
		_mrunnerdel methodrunner = mrunner;
		private Form form; // the application main form

		// constructor
		public BasicLib(Form mainform)
		{
			form = mainform;
			eventObject = new object[] { this, "" };
			b4p = SysThread.Thread.GetData(SysThread.Thread.GetNamedDataSlot("b4p"));
			if (b4p == null)
				b4p = SysThread.Thread.GetData(SysThread.Thread.GetNamedDataSlot("CRunner"));
			legacy = false;
			if (b4p.GetType().Name != "b4p")
			{
				legacy = true;
				Type b4ptype = b4p.GetType();
				miMakeRegularDictionary = b4ptype.GetMethod("MakeRegularDictionary");
				miMakeRegularVariable = b4ptype.GetMethod("MakeRegularVariable");
				miStartSub = b4ptype.GetMethod("StartSub");
				if (miMakeRegularDictionary == null)
				{
					MessageBox.Show("Incompatible Basic4ppc version!", "BasicLib");
					return;
				}
			}
			htSubs = (IDictionary)b4p.GetType().GetField("htSubs").GetValue(b4p);
			IEnumerator ienum = htSubs.Keys.GetEnumerator();
			ienum.MoveNext();
			String str = (String)ienum.Current;
			if (str.StartsWith("__"))
				prefix = "__";
			else if (str.StartsWith("_"))
				prefix = "_";
			else
				prefix = "";
			breakline = 0;
			currenttoken = _Token.ERROR;
			random = new Random();
			keywords = new Hashtable();
			keywords.Add("\n", _Token.CR);
			keywords.Add(",", _Token.COMMA);
			keywords.Add(".", _Token.DOT);
			keywords.Add(";", _Token.SEMICOLON);
			keywords.Add("+", _Token.PLUS);
			keywords.Add("-", _Token.MINUS);
			keywords.Add("&", _Token.APPEND);
			keywords.Add("^", _Token.POWER);
			keywords.Add("*", _Token.TIMES);
			keywords.Add("/", _Token.DIVIDE);
			keywords.Add("%", _Token.MOD);
			keywords.Add("(", _Token.LEFTPAREN);
			keywords.Add(")", _Token.RIGHTPAREN);
			keywords.Add("<", _Token.LT);
			keywords.Add(">", _Token.GT);
			keywords.Add("=", _Token.EQ);
			keywords.Add("<>", _Token.NE);
			keywords.Add("<=", _Token.LE);
			keywords.Add(">=", _Token.GE);
			keywords.Add("?", _Token.PRINT);
			keywords.Add(":", _Token.LABEL);
			keywords.Add("#", _Token.REM);
			keywords.Add("rem", _Token.REM);

			keywords.Add("cpi", _Token.CPI);
			keywords.Add("ce", _Token.CE);
			keywords.Add("crlf", _Token.CRLF);
			keywords.Add("ctab", _Token.CTAB);
			keywords.Add("ccasesensitive", _Token.CASESENSITIVE);
			keywords.Add("ccaseunsensitive", _Token.CASEUNSENSITIVE);
			keywords.Add("cticksperday", _Token.CTICKSPERDAY);
			keywords.Add("cticksperhour", _Token.CTICKSPERHOUR);
			keywords.Add("cticksperminute", _Token.CTICKSPERMINUTE);
			keywords.Add("ctickspersecond", _Token.CTICKSPERSECOND);
			keywords.Add("now", _Token.NOW);
			keywords.Add("version", _Token.VERSION);

			keywords.Add("date", _Token.DATE);
			keywords.Add("dateadd", _Token.DATEADD);
			keywords.Add("dateday", _Token.DATEDAY);
			keywords.Add("datedayofweek", _Token.DATEDAYOFWEEK);
			keywords.Add("datedayofyear", _Token.DATEDAYOFYEAR);
			keywords.Add("dateformat", _Token.DATEFORMAT);
			keywords.Add("datemonth", _Token.DATEMONTH);
			keywords.Add("dateparse", _Token.DATEPARSE);
			keywords.Add("dateyear", _Token.DATEYEAR);

			keywords.Add("time", _Token.TIME);
			keywords.Add("timeadd", _Token.TIMEADD);
			keywords.Add("timeformat", _Token.TIMEFORMAT); 
			keywords.Add("timehour", _Token.TIMEHOUR);
			keywords.Add("timeminute", _Token.TIMEMINUTE);
			keywords.Add("timesecond", _Token.TIMESECOND);
			keywords.Add("timeparse", _Token.TIMEPARSE);

			keywords.Add("callhost", _Token.CALLHOST);
			keywords.Add("callhostsub", _Token.CALLHOSTSUB);
			keywords.Add("callhostguisub", _Token.CALLHOSTGUISUB);
			keywords.Add("callsub", _Token.CALLSUB);
			keywords.Add("let", _Token.LET);
			keywords.Add("dim", _Token.DIM);
			keywords.Add("type", _Token.TYPE);
			keywords.Add("array", _Token.ARRAY);
			keywords.Add("getitem", _Token.GETITEM);
			keywords.Add("setitem", _Token.SETITEM);
			keywords.Add("getrank", _Token.GETRANK);
			keywords.Add("getlen", _Token.GETLEN);
			keywords.Add("gettotallen", _Token.GETTOTALLEN);
			keywords.Add("arraycopy", _Token.ARRAYCOPY);
			keywords.Add("arraylen", _Token.ARRAYLEN);
			keywords.Add("arraytotallen", _Token.ARRAYTOTALLEN);
			keywords.Add("arrayrank", _Token.ARRAYRANK);
			keywords.Add("print", _Token.PRINT);
			keywords.Add("msgbox", _Token.MSGBOX);
			keywords.Add("true", _Token.TRUE);
			keywords.Add("false", _Token.FALSE);
			keywords.Add("if", _Token.IF);
			keywords.Add("then", _Token.THEN);
			keywords.Add("else", _Token.ELSE);
			keywords.Add("elseif", _Token.ELSEIF);
			keywords.Add("end", _Token.END);
			keywords.Add("endif", _Token.ENDIF);
			keywords.Add("select", _Token.SELECT);
			keywords.Add("endselect", _Token.ENDSELECT);
			keywords.Add("sub", _Token.SUB);
			keywords.Add("endsub", _Token.ENDSUB);
			keywords.Add("for", _Token.FOR);
			keywords.Add("to", _Token.TO);
			keywords.Add("step", _Token.STEP);
			keywords.Add("next", _Token.NEXT);
			keywords.Add("do", _Token.DO);
			keywords.Add("loop", _Token.LOOP);
			keywords.Add("while", _Token.WHILE);
			keywords.Add("until", _Token.UNTIL);
			keywords.Add("exit", _Token.EXIT);
			keywords.Add("return", _Token.RETURN);
			keywords.Add("goto", _Token.GOTO);
			keywords.Add("case", _Token.CASE);
			keywords.Add("appclose", _Token.APPCLOSE);
			keywords.Add("sleep", _Token.SLEEP);
			keywords.Add("cos", _Token.COS);
			keywords.Add("sin", _Token.SIN);
			keywords.Add("tan", _Token.TAN);
			keywords.Add("acos", _Token.ACOS);
			keywords.Add("asin", _Token.ASIN);
			keywords.Add("atan", _Token.ATAN);
			keywords.Add("sqrt", _Token.SQRT);
			keywords.Add("int", _Token.INT);
			keywords.Add("abs", _Token.ABS);
			keywords.Add("max", _Token.MAX);
			keywords.Add("min", _Token.MIN);
			keywords.Add("rnd", _Token.RND);
			keywords.Add("ln", _Token.LN);
			keywords.Add("log", _Token.LOG);
			keywords.Add("round", _Token.ROUND);
			keywords.Add("not", _Token.NOT);
			keywords.Add("and", _Token.AND);
			keywords.Add("or", _Token.OR);
			keywords.Add("bitand", _Token.BITAND);
			keywords.Add("bitor", _Token.BITOR);
			keywords.Add("bitxor", _Token.BITXOR);
			keywords.Add("mod", _Token.MOD);
			keywords.Add("asc", _Token.ASC);
			keywords.Add("chr", _Token.CHR);
			keywords.Add("isdigit", _Token.ISDIGIT);
			keywords.Add("isletter", _Token.ISLETTER);
			keywords.Add("isnumber", _Token.ISNUMBER);
			keywords.Add("ispunctuation", _Token.ISPUNC);
			keywords.Add("format", _Token.FORMAT);
			keywords.Add("strat", _Token.STRAT);
			keywords.Add("strcompare", _Token.STRCOMPARE);
			keywords.Add("strindexof", _Token.STRINDEXOF);
			keywords.Add("strinsert", _Token.STRINSERT);
			keywords.Add("strjoin", _Token.STRJOIN);
			keywords.Add("strlength", _Token.STRLENGTH);
			keywords.Add("strremove", _Token.STRREMOVE);
			keywords.Add("strreplace", _Token.STRREPLACE);
			keywords.Add("strtolower", _Token.STRTOLOWER);
			keywords.Add("strtoupper", _Token.STRTOUPPER);
			keywords.Add("substring", _Token.SUBSTRING);
			keywords.Add("strsplit", _Token.STRSPLIT);
			keywords.Add("getr", _Token.GETR);
			keywords.Add("getg", _Token.GETG);
			keywords.Add("getb", _Token.GETB);
			keywords.Add("rgb", _Token.RGB);
			keywords.Add("cbeige", _Token.CBEIGE);
			keywords.Add("cblack", _Token.CBLACK);
			keywords.Add("cblue", _Token.CBLUE);
			keywords.Add("cbrown", _Token.CBROWN);
			keywords.Add("ccyan", _Token.CCYAN);
			keywords.Add("cgold", _Token.CGOLD);
			keywords.Add("cgray", _Token.CGRAY);
			keywords.Add("cgreen", _Token.CGREEN);
			keywords.Add("corange", _Token.CORANGE);
			keywords.Add("cpink", _Token.CPINK);
			keywords.Add("cpurple", _Token.CPURPLE);
			keywords.Add("cred", _Token.CRED);
			keywords.Add("csilver", _Token.CSILVER);
			keywords.Add("cwhite", _Token.CWHITE);
			keywords.Add("cyellow", _Token.CYELLOW);
			keywords.Add("*!@!@!@!*", _Token.ERROR);

			hostsubnames = new Hashtable();
	   }

	   // private event handlers

		private void BreakEvent()
		{
			if (Break != null)
			{
				eventObject[1] = "Break";
				Break(eventObject, null);
			}
		}

		private void CallHostEvent()
		{
			if (CallHost != null)
			{
				eventObject[1] = "CallHost";
				CallHost(eventObject, null);
			}
		}

		private void EndedEvent()
		{
			if (Ended != null)
			{
				eventObject[1] = "Ended";
				Ended(eventObject, null);
			}
		}

		private void LoadedEvent()
		{
			if (Loaded != null)
			{
				eventObject[1] = "Loaded";
				Loaded(eventObject, null);
			}
		}

		private void MessageEvent()
	   {
		   if (Message != null)
		   {
			   eventObject[1] = "Message";
			   Message(eventObject, null);
		   }
	   }

		private void StartingEvent()
		{
			if (Starting != null)
			{
				eventObject[1] = "Starting";
				Starting(eventObject, null);
			}
		}

		private void SteppedEvent()
		{
			if (Stepped != null)
			{
				eventObject[1] = "Stepped";
				Stepped(eventObject, null);
			}
		}


#region Basic4ppc interface

		// public methods

		// Add the Sub names contained in the array to the internal Sub call table so
		// that thay may be called directly as functions or statements rather than by using CallHostSub("whatever",...)
		public void AddSubCalls(string[] subnames)
		{
			for (int i = 0; i < subnames.Length; i++)
				this.hostsubnames.Add(subnames[i].Trim().ToLower().Replace('.', '_'), _Token.HOSTSUBCALL);
		}

		public String Call(String name, String[] paramaters)
		{
			name = name.ToLower();
			if (paramaters == null)
				paramaters = new String[0];
			if (subptrs.ContainsKey(name))
			{
				if (callstackptr < MaxStackDepth)
				{
					callstackptr++;
					localsptr++;
				}
				else
				{
					saveerror("Call : stack depth exceeded, increase the StackDepth property");
					EndedEvent();
					return "";
				}
			}
			else
			{
				saveerror("Call : Sub " + name + " not found");
				EndedEvent(); //
				return "";
			}
			localsstack[callstackptr - 1] = localvars;
			callstack[callstackptr - 1] = tokenptr;
			parameters[callstackptr - 1].Clear();
			for (int i = 0; i < paramaters.Length; i++)
				parameters[callstackptr - 1].Add(paramaters[i]);
			gotolinenum((int)subptrs[name]); // at next line NUMBER token
			// ended is true after the program has run off the inital block
			ended = false;
			_Token token;
			int stack = callstackptr - 1; // checking this allows Subs to be called in event code
			try
			{
				do
				{
					if (ended) // error might occur in the called Sub
					{
						EndedEvent();
						return "";
					}
					token = executeline(false);
				} while (token != _Token.RETURN || callstackptr != stack); // until return from event Sub
				// if this method is recursively called and we need to reset the tokenptr and currenttoken
				// when the recursion ends
				gotoplace(--tokenptr);
				return returnval;
			}
			catch (Exception e)
			{
				String msg = "Call : " + name + " - " + e.ToString() + " - " + e.Message;
				if (errorstring == "") // only return the first error
					saveerror(msg);
			}
			EndedEvent();
			return "";
		}

		public bool CanCall(String name)
		{
			return subptrs.ContainsKey(name.ToLower());
		}

		public void ClearSubCalls()
		{
			hostsubnames.Clear();
		}

		public String[] GetArray(String var)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				return (String[])globals[(int)globalvars[var]].Clone();
			else
				return new String[0];
		}

		public String GetArrayItem(String var, int index)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				return globals[(int)globalvars[var]][index];
			else
				return "";
		}

		public int GetArrayLength(String var)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				return globals[(int)globalvars[var]].Length;
			else
				return -1;
		}

		public String GetGlobal(String var)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				return globals[(int)globalvars[var]][0];
			else
				return "";
		}

		public String GetLocal(String var)
		{
			var = var.Trim().ToLower();
			if (callstackptr > 0)
			{
				if (localvars[var] != null)
					return locals[callstackptr - 1, (int)localvars[var]];
				else
					return "";
			}
			else
			{
				return GetGlobal(var);
			}
		}

		public String[] GetSubCalls()
		{
			string[] subs = new string[hostsubnames.Count];
			int i = 0;
			foreach(string sub in hostsubnames.Keys)
			{
				subs[i++] = sub;            

			}
			Array.Sort(subs);
			return subs;
		}

		public void Halt()
		{
			ended = true;
		}

		public bool LoadCodeAsArray(String[] program)
		{
			List<int> srclines = new List<int>();
			List<int> loadlines = new List<int>();
			srclines.Add(0);
			loadlines.Add(0); // first real line index is 1
		   
			int linenumber = 1;
			String line;
			subptrs = new Hashtable(); // key - string Sub name : value int linenumber
			// Char[] crlf = { '\r', '\n', ' ' };
			Char[] crlfsp = new Char[3];
			crlfsp[0] = '\r';
			crlfsp[1] = '\n';
			crlfsp[2] = ' '; // longwinded but avoids <private implementation> showing in B4ppc
			StringBuilder sb = new StringBuilder();
			for (int i = 0 ; i < program.Length ; i++)
			{
				line = program[i].Trim(crlfsp) + "\r\n"; // important SUB name followed by \r
				program[i] = line; // tidy up source, LoadCodeAsString leaves \n at start of each element
				if (line.Length > 3 && line.Substring(0, 3).ToLower() == "sub")
				{
					int j = 2;
					do
					{
						++j;
					} while (line[j] == ' ' || line[j] == '\t');
					int k = j;
					do
					{
						++k;
					} while (line[k] != '\r' && line[k] != '(');
					subptrs.Add(line.Substring(j, k - j).ToLower(), linenumber);
				}
				sb.Append(linenumber);
				sb.Append(" ");
				sb.Append(line);
				srclines.Add(linenumber);
				loadlines.Add(linenumber);
				linenumber++;
			}
			srclines.Add(linenumber);  // padding for safety
			loadlines.Add(linenumber); // padding for safety
			sourcelines = srclines.ToArray();
			loadedlines = loadlines.ToArray();
			sb.Append((char)0);
			bool success;
			source = sb.ToString();
			try
			{
				success = scanprogram();
			}
			catch (Exception e)
			{
				success = false;
				if (errorstring == "") // only return the first error
					errorstring = e.Message;
			}

			LoadedEvent();
			return success;
		}

		public bool LoadCodeAsString(String program)
		{
			int len = program.Length;
			sourcearray = program.Split(new Char[] { '\r' });
			return LoadCodeAsArray(sourcearray);
		}

		// splitting of multiple statements using the " : " character sequence as a separator
		// the sequence " : " cannot be used in a literal string code "Fred : 11" as "Fred :" & " 11"
		// error line numbers will be the new source line number but OriginalProgramLineNumber will return the original
		public bool LoadCodeAsString2(String program)
		{
			List<String> src = new List<String>();
			List<int> srclines = new List<int>();
			List<int> loadlines = new List<int>();
			srclines.Add(0);
			loadlines.Add(0); // first real line index is 1

			int len = program.Length;
			int loadline, srcline ;
			srcline = loadline = 1;
		   
			sourcearray = program.Split(new String[] { "\r\n" }, StringSplitOptions.None);
			String[] temp;
			foreach (String s in sourcearray)
			{
				loadlines.Add(loadline);
				temp = s.Split(new String[] { " : " }, StringSplitOptions.None);
				foreach (String t in temp)
				{
					src.Add(t);
					srclines.Add(srcline);
					loadline++;
				}
				srcline++;
			}
			srclines.Add(srcline); // padding for safety
			loadlines.Add(loadline); // padding for safety
			sourcearray = src.ToArray(); // save source now converted to single statements
			bool ret =  LoadCodeAsArray(sourcearray);
			sourcelines = srclines.ToArray(); // LoadCodeAsArray has set these 
			loadedlines = loadlines.ToArray();
			return ret;
		}

		public int LoadedLineNumber(int sourcelinenumber)
		{
			return loadedlines[sourcelinenumber];
		}

		public String ProgramLine(int loadedlinenumber)
		{
			if (loadedlinenumber > 0 && loadedlinenumber < (sourcearray.Length + 1))
				return sourcearray[loadedlinenumber - 1];
			else
				return "Null";
		}

		public bool Run(String[] args)
		{
			initvars(args);
			StartingEvent();
			try
			{
				runprogram();
			} 
			catch (Exception e)            
			{
				if (errorstring == "") // only return the first error
					errorstring = e.Message;
			
			}
			EndedEvent();
			return (errorstring == "" ? true : false);
		}

		public void SetArray(String var, String[] vals)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
			{
				globals[(int)globalvars[var]] = (String[])vals.Clone();
				globalsrank[(int)globalvars[var]] = new int[] { vals.Length, 0, 0 };
			}
		}

		public void SetArrayItem(String var, int index, String val)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				globals[(int)globalvars[var]][index] = val;
		}

		public void SetGlobal(String var, String val)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
				globals[(int)globalvars[var]][0] = val;
		}

		public void SetLocal(String var, String val)
		{
			var = var.Trim().ToLower();
			if (callstackptr > 0)
			{
				if (localvars[var] != null)
					locals[callstackptr - 1, (int)localvars[var]] = val;
			}
			else
			{
				SetGlobal(var, val);
			}
		}

		public void ShareArray(String var, String[] vals)
		{
			var = var.Trim().ToLower();
			if (globalvars[var] != null)
			{
				globals[(int)globalvars[var]] = vals;
				globalsrank[(int)globalvars[var]] = new int[] { vals.Length, 0, 0 };
			}
	   }
		
		public int SourceLineNumber(int loadedlinenumber)
		{
			return sourcelines[loadedlinenumber];
		}



		// public properties

		public int BreakAtLine
		{ get { return breakline; } set { breakline = value; } }

		public double DllVersion
		{ get { return version; } }
 
		public String ErrorString
		{ get { return errorstring; } }

		public bool Finished
		{ get { return ended; } }

		public int LineNumber
		{ get { return linenum; } }

		public String MessageString
		{ get { return messagestring; } }

		public String[] Program
		{ get { return sourcearray; } }

		public int ProgramLines
		{ get { return sourcearray.Length; } }

		public int StackDepth
		{ get { return MaxStackDepth; }
			set
			{
				MaxStackDepth = value;
			}
		}

		public bool Step
		{ get { return stepping ; } set { stepping = value; } }

		public String SubjectString
		{ get { return subjectstring; } }


		// public development methods

		public string devGlobal(int id)
		{
			int[] values = new int[globalvars.Count];
			string[] keys = new string[globalvars.Count];
			globalvars.Values.CopyTo(values, 0);
			globalvars.Keys.CopyTo(keys, 0);
			string global = "";
			for (int i = 0; i < values.Length; i++)
			{
				if (values[i] == id)
				{
					global = keys[i];
					break;
				}
			}
			return global;
		}

		public string devLiteral(int id)
		{
			return (String)literals[id];
		}

		public string devLocal(string sub, int id)
		{
			sub = sub.ToLower();
			Hashtable localvars = (Hashtable)subvars[sub];
			int[] values = new int[localvars.Count];
			string[] keys = new string[localvars.Count];
			localvars.Values.CopyTo(values, 0);
			localvars.Keys.CopyTo(keys, 0);
			string local = "";
			for (int i = 0; i < values.Length; i++)
			{
				if (values[i] == id)
				{
					local = keys[i];
					break;
				}
			}
			return local;
		}

		public string[] devLocals(String sub)
		{
			sub = sub.Trim().ToLower();
			Hashtable localsh = (Hashtable)localvars[sub];
			if (localvars[sub] != null)
			{
				string[] locals = new string[localsh.Count];
				localsh.CopyTo(locals, 0);
				return locals;            
			}
			else
				return new string[0];
		}

		public string devTokenName(int tokenvalue)
		{
			return ((_Token)tokenvalue).ToString();
		}
		
		public string devTokenNameAt(int index)
		{
			string token= ((_Token)source[index]).ToString().ToLower();
			return token.Substring(0, 1).ToUpper() + token.Substring(1);
		}

		public int devTokenValue(string tokenname)
		{
			return (int)Enum.Parse(typeof(_Token), tokenname, true);
		}

		public int devTokenValueAt(int index)
		{
			return source[index];
		}


		// public development properties

		public string devCurrentSubName
		{ get {
				int[] linenums = new int[subptrs.Count];
				string[] subs = new string[subptrs.Count];
				subptrs.Values.CopyTo(linenums, 0);
				subptrs.Keys.CopyTo(subs, 0);
				string sub = "";
				for (int i = 0; i < linenums.Length; i++)
				{
					if (linenums[i] >= linenum)
					{
						sub = subs[i];
						break;
					}
				}
				return sub; }
		}

		public string[] devGlobals
		{
			get
			{
				string[] globals = new string[globalvars.Count];
				subptrs.CopyTo(globals, 0);
				return globals;
			}
		}

		public string[] devSubNames
		{
			get
			{
				string[] subs = new string[subptrs.Count];
				subptrs.CopyTo(subs, 0);
				return subs;
			}
		}

		public String devTokens
		{ get { return source; } }

		public int devTokenCount
		{ get { return source.Length; } }


		//private methods
		private String b4pname(String name)
		{
			if (name.IndexOf('.') >= 0)
				return prefix + name.ToLower(cul).Replace(".", "_");
			//return prefix + "main_" + name.ToLower(cul); removed to allow AddSubCalls to work
			return prefix + name.ToLower(cul); 
		}

		private static object mrunner(object ctl, MethodInfo method, object[] args)
		{
			return method.Invoke(ctl, args);
		}

		private string chsIDE(string subname, string[] parameters, bool invoke)
		{
			object sub = htSubs[b4pname(subname)]; // gets a Sub object
			if (sub == null)
			{
				saveerror("CallHostSub : Sub " + subname + " not found");
				return "";
			}
			Type returnType = (Type)sub.GetType().GetField("returnType").GetValue(sub);
			if (returnType != typeof(string))
			{
				saveerror("CallHostSub : Sub " + subname + " must return a String");
				return "";
			}
			object subplace = sub.GetType().GetField("place").GetValue(sub);
			List<string> varsNames = (List<string>)sub.GetType().GetField("varsNames").GetValue(sub); // Place
			List<Type> ptlist = (List<Type>)sub.GetType().GetField("parametersTypes").GetValue(sub);
			int numparams = varsNames.Count;
			if (numparams != parameters.Length)
			{
				string param = numparams + " parameter";
				if (numparams != 1)
					param += "s";
				saveerror("CallHostSub : Sub " + subname + " requires " + numparams + " parameters");
				return "";
			}
			IDictionary localsReg = (IDictionary)miMakeRegularDictionary.Invoke(b4p, null);
			if (numparams > 0)
			{
				for (int i = 0; i < numparams; i++)
				{
					if (ptlist[i] != typeof(string))
					{
						saveerror("CallHostSub : Sub " + sub + " parameter " + varsNames[i] + " must be of type String");
						return "";
					}
					object regstr = miMakeRegularVariable.Invoke(b4p, new object[] { typeof(string) });
					FieldInfo fireg = regstr.GetType().GetField("value", BindingFlags.NonPublic | BindingFlags.Instance);
					fireg.SetValue(regstr, parameters[i]);
					localsReg.Add(varsNames[i], regstr);
				}
			}
			object[] oargs = new object[] { subplace, localsReg, null };
			object resultReg;
			if (invoke)
			{
				object[] oa = new object[] { b4p, miStartSub, oargs };
				resultReg = form.Invoke(methodrunner, oa);
			}
			else
			{
				resultReg = miStartSub.Invoke(b4p, oargs);
			}
			return (string)resultReg.GetType().GetField("value", BindingFlags.NonPublic | BindingFlags.Instance).GetValue(resultReg);

		}






		// chs0 is a separate sub to stop the CLR JITing the delegate lookup
		// it seems to do this only if the sub is run and then errors looking for 1.exe
		// in a separate sub it is not run so not parsed by the runtime
		private string CallHostSub0(string sub, bool invoke)
		{
			if (legacy)
				return (string) chsIDE(sub, new string[0], invoke);
			return chs0(sub, invoke);
		}

		private string chs0(string sub, bool invoke)
		{
			b4p.del0 del = (b4p.del0)htSubs[b4pname(sub)];
			if (del == null)
			{
				saveerror("CallHostSub : Sub " + sub + " not found or wrong number of parameters");
				return "";
			}
			if (invoke)
				return (string) form.Invoke(del);
			else
				return del();
		}


		// chs1 is a separate sub to stop the CLR JITing the delegate lookup
		// it seems to do this only if the sub is run and then errors looking for 1.exe
		// in a separate sub it is not run so not parsed by the runtime
		private string CallHostSub1(string sub, string a, bool invoke)
		{
			if (legacy)
				return (string)chsIDE(sub, new string[] { a }, invoke);
			return chs1(sub, a, invoke);
		}

		private string chs1(string sub, string a, bool invoke)
		{
			b4p.del1 del = (b4p.del1)htSubs[b4pname(sub)];
			if (del == null)
			{
				saveerror("CallHostSub : Sub " + sub + " not found");
				return "";
			}
			if (invoke)
			{
				object[] args = new object[1] { a };
				return (string) form.Invoke(del, args);
			}
			else
				return del(a);
		}

		// chs2 is a separate sub to stop the CLR JITing the delegate lookup
		// it seems to do this only if the sub is run and then errors looking for 1.exe
		// in a separate sub it is not run so not parsed by the runtime
		private string CallHostSub2(string sub, string a, string b, bool invoke)
		{
			if (legacy)
				return (string)chsIDE(sub, new string[] { a, b }, invoke);
			return chs2(sub, a, b, invoke);
		}

		private string chs2(string sub, string a, string b, bool invoke)
		{
			// minimise locked time by looking up first
			b4p.del2 del = (b4p.del2)htSubs[b4pname(sub)];
			if (del == null)
			{
				saveerror("CallHostSub : Sub " + sub + " not found");
				return "";
			}
			if (invoke)
			{
				object[] args = new object[2] { a, b };
				return (string) form.Invoke(del, args);
			}
			else
				return del(a, b);
		}

		// chs3 is a separate sub to stop the CLR JITing the delegate lookup
		// it seems to do this only if the sub is run and then errors looking for 1.exe
		// in a separate sub it is not run so not parsed by the runtime
		private string CallHostSub3(string sub, string a, string b, string c, bool invoke)
		{
			if (legacy)
				return (string)chsIDE(sub, new string[] { a, b, c }, invoke);
			return chs3(sub, a, b, c, invoke);
		}

		private string chs3(string sub, string a, string b, string c, bool invoke)
		{
			// minimise locked time by looking up first
			b4p.del3 del = (b4p.del3)htSubs[b4pname(sub)];
			if (del == null)
			{
				saveerror("CallHostSub : Sub " + sub + " not found");
				return "";
			}
			if (invoke)
			{
				object[] args = new object[3] { a, b, c} ;
				return (string) form.Invoke(del, args);
			}
			else
				return del(a, b, c);
		}





#endregion Basic4ppc interface


#region Interpreter

		// execution
		private String getvariable()
		{
			int substart = tokenptr;
			do
			{
				tokenptr++;
			} while (Char.IsLetterOrDigit(source[tokenptr])
				|| source[tokenptr] == '_'
				|| source[tokenptr] == '#');
			int length = tokenptr - substart;
			nextptr = tokenptr;
			return source.Substring(substart, length).ToLower(); //.TrimStart(new char[] { ' ' });
		}

		private bool addvariable(bool insub, bool parameter, StringBuilder program)
		{
			String varname = getvariable();
			int index;
			if (insub)
			{
				if (globalvars.ContainsKey(varname))
				{
					if (parameter)
					{
						saveerror("Local variable cannot hide global variable " + varname);
						return false;
					}
					index = (int)globalvars[varname];
					program.Append('G');
					program.Append((Char)index);
				}
				else
				{
					if (!localvars.ContainsKey(varname))
						localvars.Add(varname, localvars.Count);
					index = (int)localvars[varname];
					program.Append('L');
					program.Append((Char)index);
				}
			}
			else // global
			{
				if (!globalvars.ContainsKey(varname))
					globalvars.Add(varname, globalvars.Count);
				index = (int)globalvars[varname];
				program.Append('G');
				program.Append((Char)index);
			}
			return true;
		}

		private char addliteral(string literal)
		{
			int idx = literals.IndexOf(literal);
			if (idx < 0)
			{
				idx = literals.Count;
				literals.Add(literal);

			}
			return (char)idx;
		}

		private bool scanprogram()
		{
			StringBuilder program = new StringBuilder();
			globalvars = new Hashtable(); // key - string name : value int index
			subvars = new Hashtable(); // key - int linenumber : value hashtable localvars
			lineptrs = new Hashtable(); // key - int linenumber : value int linenums index
			labelptrs = new Hashtable(); // key - string name : value int linenum
			statements = new ArrayList(); // the statements in linenumber order
			statements.Add(_Token.NULL); // statements index  = linenumber
			literals = new ArrayList(); // literal strings used by the program   
			bool insub = false; // true while processing a sub
			bool parameter = false; // true while processing a parameter variable
			String currentsub = ""; // holds name of last Sub statement
			globalvars.Add("args", 0); // globals[0] is args
			errorstring = "";
			maxlocals = 0;
			tokeniserinit();
			do
			{
				// add linenumber
				linenum = (int)double.Parse(source.Substring(tokenptr, nextptr - tokenptr), cul);;
				lineptrs.Add(linenum, program.Length);
				program.Append((Char)linenum);
				fetchnextoken();

				// add first token of line
				if (currenttoken == _Token.ENDSUB)
				{
					insub = false;
					subvars.Add(currentsub, localvars);
					maxlocals = Math.Max(maxlocals, localvars.Count);
					currenttoken = _Token.ENDSUB;
				}                
				if (currenttoken == _Token.END)
				{
					fetchnextoken(); // should be IF, SELECT or SUB
					if (currenttoken == _Token.IF)
					{
						currenttoken = _Token.ENDIF;
					}
					if (currenttoken == _Token.SUB)
					{
						insub = false; 
						subvars.Add(currentsub, localvars);
						maxlocals = Math.Max(maxlocals, localvars.Count);
						currenttoken = _Token.ENDSUB;
					}
					if (currenttoken == _Token.SELECT)
						currenttoken = _Token.ENDSELECT;
				}
				if (currenttoken == _Token.VARIABLE)
				{
					int oldptr = tokenptr;
					int oldnext = nextptr;
					String labelname = getvariable();
					fetchnextoken();
					if (currenttoken == _Token.LABEL)
					{
						fetchnextoken(); // skip to next token
						labelptrs.Add(labelname, linenum);
					}
					else
					{
						tokenptr = oldptr;
						nextptr = oldnext;
						currenttoken = _Token.VARIABLE;
					}
				}
				if (currenttoken == _Token.ELSE)
				{
					fetchnextoken(); // might be IF
					if (currenttoken == _Token.IF)
					{
						currenttoken = _Token.ELSEIF;
						statements.Add(_Token.ELSEIF);
					}
					else
					{
						program.Append((Char)_Token.ELSE);
						statements.Add(_Token.ELSE);
					}
				}
				else
				{
					statements.Add(currenttoken); // save first token of line
				}
				program.Append((Char)currenttoken);
				if (currenttoken == _Token.DO ||
					currenttoken == _Token.FOR ||
					currenttoken == _Token.SELECT ||
					currenttoken == _Token.CASE ||
					currenttoken == _Token.ELSEIF ||
					currenttoken == _Token.IF ||
					currenttoken == _Token.GOTO)
				{
					// placeholder for exit or else/elseif line number
					program.Append('x');
					if (currenttoken == _Token.IF ||
						currenttoken == _Token.SELECT ||
						currenttoken == _Token.GOTO)
						// placeholder for endif line number
						program.Append('x'); 
				}
				parameter = false;
				if (currenttoken == _Token.ERROR)
				{
					return false;
				}
				if (currenttoken == _Token.REM)
				{
					// REM is already appended to program
					char chr;
					// ignore rest of line til CR
					do
					{
						chr = source[++tokenptr];
					} while (chr != '\r');
					nextptr = tokenptr;
					// back to normal token processing
					fetchnextoken(); 
					program.Append((Char)currenttoken); // should be CR
				}
				if (currenttoken != _Token.CR) // allows blank lines
				{
					if (currenttoken == _Token.SUB)
					{
						insub = true;
						localvars = new Hashtable();
						fetchnextoken(); // at start of sub name
						int substart = tokenptr;
						do
						{
							tokenptr++;
						} while (source[tokenptr] != '\r' && source[tokenptr] != '(');
						int length = tokenptr - substart;
						subname = source.Substring(substart, length).ToLower().TrimStart(new char[] { ' ' });
						nextptr = tokenptr;
						program.Append(addliteral(subname));
						currentsub = subname;
						parameter = true;
					}
					else if (currenttoken == _Token.VARIABLE)
					{
						if (!addvariable(insub, parameter, program))
						{
							return false;
						}
					}
					else if (currenttoken == _Token.GOTO)
					{
						fetchnextoken(); // at start of label name
						String label = getvariable();
						program.Append((Char)label.Length);                        
						program.Append(label);
						if (keywords.ContainsKey(label))
						{
							saveerror("Goto : keyword " + label + " cannot be used as label");
							return false;
						}
						if (subptrs.ContainsKey(label))
						{
							saveerror("Goto : cannot use Sub name " + label + " as a label");
							return false;
						}
					}
					else if (currenttoken == _Token.CALL || currenttoken == _Token.HOSTSUBCALL)
					{
						program.Append(addliteral(subname));
					}

					// process rest of line
					if (currenttoken != _Token.CR && currenttoken != _Token.ERROR) // allows blank lines
					{
						do
						{
							fetchnextoken();
							program.Append((Char)currenttoken);
							if (currenttoken == _Token.NUMBER)
							{
								String numstr = source.Substring(tokenptr, nextptr - tokenptr);
								program.Append(addliteral(numstr));
							}
							else if (currenttoken == _Token.STRING)
							{
								String str = source.Substring(tokenptr + 1, nextptr - tokenptr - 2);
								program.Append(addliteral(str));
							}
							else if (currenttoken == _Token.VARIABLE)
							{
								if (!addvariable(insub, parameter, program))
								{
									return false;
								}
							}
							else if (currenttoken == _Token.CALL || currenttoken == _Token.HOSTSUBCALL)
							{
								program.Append(addliteral(subname));
							}
						} while (currenttoken != _Token.CR &&
							currenttoken != _Token.ENDOFINPUT &&
							currenttoken != _Token.ERROR);
					}                   
				}
				if (currenttoken == _Token.CR)
				{
					fetchnextoken(); // NUMBER or ENDOFINPUT
				}
			} while (currenttoken != _Token.ENDOFINPUT && currenttoken != _Token.ERROR);
			if (currenttoken == _Token.ENDOFINPUT)
			{
				program.Append((char)(linenum + 1));
				program.Append((char)_Token.SUB); // Sub statement ends program
				bool success = scanagain(program);
				source = program.ToString(); // set source for Tokens property
				if (success)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				source = program.ToString(); // set source for Tokens property
				return false;
			}
		}

		private bool scanagain(StringBuilder program)
		{
			for (int i = 0; i < statements.Count; i++)
			{
				linenum = i; // for error reports
				_Token intoken = (_Token)statements[i];
				if (intoken == _Token.DO || intoken == _Token.FOR)
				{
					_Token outtoken = _Token.LOOP;
					if (intoken == _Token.FOR)
						outtoken = _Token.NEXT;
					int target = findmatchingtoken(intoken, outtoken, i, 1);
					int place = (int)lineptrs[i];
					program[place + 2] = (char)target;
					if (target < 0)
					{
						 return false; // error not found
					}
				}
				if (intoken == _Token.ELSEIF || intoken == _Token.IF)
				{
					int elseline, endifline;
					findelseline(i, out elseline, out endifline);
					int place = (int)lineptrs[i];
					program[place + 2] = (char)elseline;
					if (intoken == _Token.IF)
						program[place + 3] = (char)endifline;
				}
				if (intoken == _Token.SELECT || intoken == _Token.CASE)
				{
					int caseline, endselectline;
					findcaseline(i, out caseline, out endselectline);
					int place = (int)lineptrs[i];
					program[place + 2] = (char)caseline;
					if (intoken == _Token.SELECT)
						program[place + 3] = (char)endselectline;
				}
				if (intoken == _Token.GOTO)
				{
					int forstackptr, ifstackptr;
					int place = (int)lineptrs[i];
					int strlen = program[place + 4];
					string label = program.ToString(place + 5, strlen);
					if (!checkgotoscope(i, label, out forstackptr, out ifstackptr))
						return false;
					program[place + 2] = (char)forstackptr;
					program[place + 3] = (char)ifstackptr;

				}
			}
			return true;
		}

		private void findelseline(int linenum, out int elseline, out int endifline)
		{
			int stackptr = 0;
			elseline = 0;
			endifline = 0;
			for (int i = linenum + 1; i < statements.Count; i++)
			{
				if (_Token.IF == (_Token)statements[i])
				{
					stackptr++;
				}
				if ( (_Token.ELSE == (_Token)statements[i] || _Token.ELSEIF == (_Token)statements[i]) &&
						elseline == 0 &&
						stackptr == 0) // found first ELSE of possible Else Ifs
				{
					elseline = i;
				}                
				if (_Token.ENDIF == (_Token)statements[i])
				{
					if (stackptr > 0)
					{
						stackptr--;
					}
					else
					{
						endifline = i;
						break;
					}
				}
				if (_Token.ENDSUB == (_Token)statements[i])
				{
					break;
				}
			}
		}

		private int findmatchingtoken(_Token intoken, _Token outtoken, int startline, int lineoffset)
		{
			int index = startline;
			int stackptr = -1;// startline should point to intoken
			for (int i = index; i < statements.Count; i++)
			{
				if (intoken == (_Token)statements[i])
				{
					stackptr++;
				}
				if (outtoken == (_Token)statements[i])
				{
					if (stackptr > 0)
					{
						stackptr--;
					}
					else
					{
						if ((i + lineoffset) < statements.Count)
							return i + lineoffset;
						else
						{
							saveerror("No program line to execute after " + outtoken);
							return -1; // target not in program
						}
					}
				}
				if (_Token.ENDSUB == (_Token)statements[i])
				{
					break;
				}
			}
			saveerror("Matching " + outtoken + " for " + intoken + " not found");
			return -1;
		}

		private void findcaseline(int linenum, out int caseline, out int endselectline)
		{
			caseline = 0;
			endselectline = 0;
			for (int i = linenum + 1; i < statements.Count; i++)
			{
				if (_Token.CASE == (_Token)statements[i])
				{
					if (caseline == 0)
						caseline = i; // first CASE
				}
				if (_Token.ENDSELECT == (_Token)statements[i])
				{
					endselectline = i;
					break;
				}
				if (_Token.ENDSUB == (_Token)statements[i])
				{
					break;
				}
			}
		}

		private bool checkgotoscope(int linenum, string label, out int forstackptr, out int ifstackptr)
		{
			int lastscopechange = 0;
			forstackptr = 0;
			ifstackptr = 0;
			object labellinenum = labelptrs[label];
			if (labellinenum != null)
			{
				bool insub = true;
				bool scopeok = true;
				int scopeptr = 0;
				if ((int)labellinenum > linenum)
				{ // going forwards
					for (int i = linenum + 1; i < (int)labellinenum; i++)
					{
						if ((_Token)statements[i] == _Token.FOR || (_Token)statements[i] == _Token.DO)
						{
							forstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.IF)
						{
							ifstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.SELECT)
						{
							ifstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.NEXT || (_Token)statements[i] == _Token.LOOP)
						{
							forstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.ENDIF)
						{
							ifstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.ENDSELECT)
						{
							ifstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.ENDSUB)
						{
							insub = false;
							break;
						}
					}
				}
				else
				{ // going backwards
					for (int i = linenum - 1; i > (int)labellinenum; i--)
					{
						if ((_Token)statements[i] == _Token.FOR || (_Token)statements[i] == _Token.DO)
						{
							forstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.IF)
						{
							ifstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.SELECT)
						{
							ifstackptr--;
							lastscopechange = -1;
							scopeptr -= 1;
						}
						else if ((_Token)statements[i] == _Token.NEXT || (_Token)statements[i] == _Token.LOOP)
						{
							forstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.ENDIF)
						{
							ifstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.ENDSELECT)
						{
							ifstackptr++;
							lastscopechange = 1;
							scopeptr += 1;
						}
						else if ((_Token)statements[i] == _Token.SUB)
						{
							insub = false;
							break;
						}
					}
				}
				if (lastscopechange == 1 ||
					(lastscopechange == -1 && scopeptr > 0))
					scopeok = false;
				if (insub)
				{
					if (!scopeok)
					{
						saveerror("Goto : Label " + label + " is in illegal scope");
						return false;
					}
				}
				else
				{
					saveerror("Goto : label " + label + " is outside Sub");
					return false;
				}
				return true;
			}
			else
			{
				saveerror("Goto : label " + label + " not found");
				return false;
			}
		}

		private void saveerror(String msg)
		{
			{
				ended = true;
				if (errorstring == "") // show only first error message
					errorstring = msg;
			}
		}

		private void runprogram()
		{
			do
			{
				if (ended)
				{
					return;
				}
				executeline(false);
			} while (true);
		}

		private void gotolinenum(int linenum)
		{
			try
			{
				int place = (int)lineptrs[linenum];
				tokenptr = place - 1; 
				getnexttoken(); // get token/linenumber at place
			}
			catch
			{
				saveerror("Goto : line number " + linenum + " not found");
			}
		}

		private void gotoplace(int place)
		{
				tokenptr = place - 1;
				getnexttoken(); // point to token at place
		}

		private _Token executeline(bool infactor)
		{
			linenum = (int)currenttoken;
			getnexttoken();
			if (stepping)
				SteppedEvent();
			if (linenum == breakline)
				BreakEvent();
			_Token token = statement(infactor);
			return token;
		}

		// variables

		private void initvars(String[] args)
		{
			selectvals = new String[MaxStackDepth + 1]; // the +1s are used at main code level by [callstackptr]
			endselectlines = new int[MaxStackDepth + 1];
			matchedcases = new bool[MaxStackDepth + 1]; 
			forstack = new _forstruct[MaxStackDepth, MaxStackDepth + 1];
			ifstack = new _ifstruct[MaxStackDepth, MaxStackDepth + 1];
			ifstackptr = new int[MaxStackDepth + 1];
			forstackptr = new int[MaxStackDepth + 1];
			for (int i = 0; i <= MaxStackDepth; i++)
				ifstackptr[i] = forstackptr[i] = 0;
			callstack = new int[MaxStackDepth]; // only used at Sub code levels by [callstackptr-1]
			localsstack = new Hashtable[MaxStackDepth];
			locals = new String[MaxStackDepth, maxlocals]; 
			parameters = new ArrayList[MaxStackDepth];
			for (int i = 0; i < MaxStackDepth; i++)
				parameters[i] = new ArrayList();
			globals = new String[globalvars.Count][];
			globalsrank = new int[globalvars.Count][];
			for (int i = 0; i < globalvars.Count; i++)
			{
				globals[i] = new string[] { "" };
				// [0] is rightmost, [1] is middle, [2] is leftmost dimension
				globalsrank[i] = new int[] { 1, 0, 0 };
			}
			dateformat = "dd/MM/yyyy";
			timeformat = "HH:mm";
			if (args.Length > 0)
			{
				globals[0] = args;
				globalsrank[0][0] = args.Length;
			}
			callstackptr = 0;
			localsptr = 0;
			tokendispatcherinit();
			errorstring = "";
			subjectstring = "";
			messagestring = "";
			ended = false;
		}
	   
		private string getvariable(int[] var)
		{
			if (var[1] < 0) // if a global
			{
				int [] dims = globalsrank[var[0]];
				// [2] is rightmost, [3] is middle, [4] is leftmost dimension
				if (var[3] != 0) // only check multi-dimensional array access
				{
					if ((var[2] >= dims[0]) ||
						(dims[1] > 0 && var[3] >= dims[1]) ||
						(dims[2] > 0 && var[4] >= dims[2]))
						saveerror("An index was greater than a dimension of the multi-dimensional array.");
				}
				int index = dims[0] * (dims[1] * var[4] + var[3]) + var[2];
				return globals[var[0]][index];
			}
			else
				return locals[localsptr - 1, var[0]];
		}

		private void setvariable(int[] var, string value)
		{
			if (var[1] < 0) // if a global
			{
				int[] dims = globalsrank[var[0]];
				// [2] is rightmost, [3] is middle, [4] is leftmost dimension
				if (var[3] !=0) // only check multi-dimensional array access
				{
					if ((var[2] >= dims[0]) ||
						(dims[1] > 0 && var[3] >= dims[1]) ||
						(dims[2] > 0 && var[4] >= dims[2]))
						saveerror("An index was greater than a dimension of the multi-dimensional array.");
				}
				int index = dims[0] * (dims[1] * var[4] + var[3]) + var[2];
				globals[var[0]][index] = value;
			}
			else
				locals[localsptr - 1, var[0]] = value;
		}

		// expressions
		
		private string factor()
		{
			int p, q;
			int[] var, lkup;
			Color color;
			string r, s, t, u; 
			switch (currenttoken)
			{
				case _Token.NUMBER:
					r = tokenisernum().ToString(cul);
					getnexttoken();
					break;
				case _Token.TRUE:
					r = "true";
					getnexttoken();
					break;
				case _Token.FALSE:
					r = "false";
					getnexttoken();
					break;
				case _Token.CPI:
					r = Math.PI.ToString(cul);
					getnexttoken();
					break;
				case _Token.CE:
					r = Math.E.ToString(cul);
					getnexttoken();
					break;
				case _Token.CRLF:
					r = "\r\n";
					getnexttoken();
					break;
				case _Token.CTAB:
					r = "\t";
					getnexttoken();
					break;
				case _Token.CASESENSITIVE:
					r = "0";
					getnexttoken();
					break;
				case _Token.CASEUNSENSITIVE:
					r = "1";
					getnexttoken();
					break;
				case _Token.CTICKSPERDAY:
					r = TimeSpan.TicksPerDay.ToString(cul);
					getnexttoken();
					break;
				case _Token.CTICKSPERHOUR:
					r = TimeSpan.TicksPerHour.ToString(cul);
					getnexttoken();
					break;
				case _Token.CTICKSPERMINUTE:
					r = TimeSpan.TicksPerMinute.ToString(cul);
					getnexttoken();
					break;
				case _Token.CTICKSPERSECOND:
					r = TimeSpan.TicksPerSecond.ToString(cul);
					getnexttoken();
					break;
				case _Token.NOW:
					r = DateTime.Now.Ticks.ToString(cul);
					getnexttoken();
					break;
				case _Token.VERSION:
					r = version.ToString(cul);
					getnexttoken();
					break;
				case _Token.MINUS: // unary minus
					// make unary "-x" equivalent to "0-x"
					// -2^2 returns -4
					r = "0";
					// alternatively give unary- highest precedence
					// -2^2 returns 4
					//getnexttoken();
					//r = factor();
					//r = (-double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.PLUS: // unary plus
					// make unary "+x" equivalent to "0+x"
					// +2^2 returns -4
					r = "0";
					// alternatively give unary+ highest precedence
					//getnexttoken();
					//r = factor();
					break;
				case _Token.STRING:
					r = tokeniserliteral();
					getnexttoken();
					break;
				case _Token.LEFTPAREN:
					getnexttoken();
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Expression");
					break;
					// array function
				case _Token.ARRAYLEN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "ArrayLen");
					var = tokeniservariable();
					getnexttoken();
					r = "1";
					if (currenttoken == _Token.COMMA)
					{
						getnexttoken();
						r = expression();
					}
					checktoken(_Token.RIGHTPAREN, "ArrayLen");
					p = (int)double.Parse(r, cul) - 1;
					lkup = new int[] { 0, 2, 1 };
					if (globalsrank[var[0]][2] != 0)
						lkup = new int[] { 2, 1, 0 };
					else if (globalsrank[var[0]][1] != 0)
						lkup = new int[] { 1, 0, 2 };
					r = globalsrank[var[0]][lkup[p]].ToString(cul);
					break;
				case _Token.ARRAYTOTALLEN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "ArrayTotalLen");
					var = tokeniservariable();
					getnexttoken();
					checktoken(_Token.RIGHTPAREN, "ArrayTotalLen");
					r = globals[var[0]].Length.ToString(cul);
					break;
				case _Token.ARRAYRANK:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "ArrayRank");
					var = tokeniservariable();
					getnexttoken();
					checktoken(_Token.RIGHTPAREN, "ArrayRank");
					var = globalsrank[var[0]];
					if (var[2] != 0)
						r = "3";
					else if (var[1] != 0)
						r = "2";
					else
						r = "1";
					break;
				case _Token.GETITEM:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetItem");
					r = expression();
					checktoken(_Token.COMMA, "GetItem");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "GetItem");
					r = globals[(int)globalvars[r]][(int)double.Parse(s, cul)];
					break;
				case _Token.GETLEN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetLen");
					s = expression();
					r = "1";
					if (currenttoken == _Token.COMMA)
					{
						getnexttoken();
						r = expression();
					}
					checktoken(_Token.RIGHTPAREN, "GetLen");
					q = (int)globalvars[s];
					p = (int)double.Parse(r, cul) - 1;
					lkup = new int[] { 0, 2, 1 };
					if (globalsrank[q][2] != 0)
						lkup = new int[] { 2, 1, 0 };
					else if (globalsrank[q][1] != 0)
						lkup = new int[] { 1, 0, 2 };
					r = globalsrank[q][lkup[p]].ToString(cul);
					break;
				case _Token.GETTOTALLEN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetTotalLen");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "GetTotalLen");
					r = globals[(int)globalvars[r]].Length.ToString(cul);
					break;
				case _Token.GETRANK:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetRank");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "GetRank");
					q = (int)globalvars[r];
					var = globalsrank[q];
					if (var[2] != 0)
						r = "3";
					else if (var[1] != 0)
						r = "2";
					else
						r = "1";
					break;
				// math functions
				case _Token.COS:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Cos");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Cos");
					r = Math.Cos(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.SIN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Sin");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Sin");
					r = Math.Sin(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.TAN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Tan");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Tan");
					r = Math.Tan(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.ACOS:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Acos");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Acos");
					r = Math.Acos(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.ASIN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Asin");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Asin");
					r = Math.Asin(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.ATAN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Atan");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Atan");
					r = Math.Atan(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.SQRT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Sqrt");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Sqrt");
					r = Math.Sqrt(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.INT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Int");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Int");
					r = Math.Floor(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.ABS:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Abs");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Abs");
					r = Math.Abs(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.LN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Ln");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Ln");
					r = Math.Log(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.LOG:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Log");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Log");
					r = Math.Log10(double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.RND:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Rnd");
					r = expression();
					checktoken(_Token.COMMA, "Rnd");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "Rnd");
					r = random.Next((int)double.Parse(r, cul), (int)double.Parse(s, cul)).ToString(cul);
					break;
				case _Token.ROUND:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Round");
					r = expression();
					s = "0"; 
					if (currenttoken == _Token.COMMA)
					{
						getnexttoken();
						s = expression();
					}
					checktoken(_Token.RIGHTPAREN, "Round");
					int digits = (int)double.Parse(s, cul);
					r = Math.Round(double.Parse(r, cul), digits).ToString(cul);
					break;
				case _Token.MAX:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Max");
					r = expression();
					checktoken(_Token.COMMA, "Max");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "Max");
					r = Math.Max((double.Parse(r, cul)), double.Parse(s)).ToString(cul);
					break;
				case _Token.MIN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Min");
					r = expression();
					checktoken(_Token.COMMA, "Min");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "Min");
					r = Math.Min((double.Parse(r, cul)), double.Parse(s, cul)).ToString(cul);
					break;
				case _Token.NOT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Not");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Not");
					r = (r == "true" ? "false" : "true");
					break;
					// string functions
				case _Token.ASC:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Asc");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Asc");
					r = ((int)r[0]).ToString(cul);
					break;
				case _Token.CHR:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Chr");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Chr");
					r = ((char)double.Parse(r, cul)).ToString(cul);
					break;
				case _Token.FORMAT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Format");
					r = expression();
					checktoken(_Token.COMMA, "Format");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "Format");
					if (s.ToLower()[0] == 'd' || s.ToLower()[0] == 'x')
					{ r = ((int)double.Parse(r, cul)).ToString(s, cul); }
					else { r =  double.Parse(r, cul).ToString(s, cul); }
					break;
				case _Token.ISDIGIT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "IsDigit");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "IsDigit");
					r = char.IsDigit(r[0]).ToString(cul).ToLower(cul);
					break;
				case _Token.ISLETTER:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "IsLetter");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "IsLetter");
					r = char.IsLetter(r[0]).ToString(cul).ToLower(cul);
					break;
				case _Token.ISNUMBER:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "IsNumber");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "IsNumber");
					r = char.IsNumber(r[0]).ToString(cul).ToLower(cul);
					break;
				case _Token.ISPUNC:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "IsPunc");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "IsPunc");
					r = char.IsPunctuation(r[0]).ToString(cul).ToLower(cul);
					break;
				case _Token.STRAT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrAt");
					r = expression();
					checktoken(_Token.COMMA, "StrAt");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "StrAt");
					r = r[(int) double.Parse(s, cul)].ToString(cul);
					break;
				case _Token.STRCOMPARE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrCompare");
					r = expression();
					checktoken(_Token.COMMA, "StrCompare");
					s = expression();
					t = "1"; //case insensitive
					if (currenttoken == _Token.COMMA)
					{
						getnexttoken();
						t = expression();
					}
					checktoken(_Token.RIGHTPAREN, "StrCompare");
					if (double.Parse(t, cul) == 0)
					{
						r = string.Compare(r, s, false, cul).ToString(cul);
					}
					else
					{
						r = string.Compare(r, s, true, cul).ToString(cul);
					}
					break;
				case _Token.STRINDEXOF:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrIndexOf");
					r = expression();
					checktoken(_Token.COMMA, "StrIndexOf");
					s = expression();
					checktoken(_Token.COMMA, "StrIndexOf");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "StrIndexOf");
					r = r.IndexOf(s, (int)double.Parse(t, cul)).ToString(cul);
					break;
				case _Token.STRINSERT:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrInsert");
					r = expression();
					checktoken(_Token.COMMA, "StrInsert");
					s = expression();
					checktoken(_Token.COMMA, "StrInsert");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "StrInsert");
					r = r.Insert((int)double.Parse(s, cul), t).ToString(cul);
					break;
				case _Token.STRJOIN:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrJoin");
					var = tokeniservariable();
					getnexttoken();
					checktoken(_Token.COMMA, "StrJoin");
					s = expression();
					checktoken(_Token.RIGHTPAREN, "StrJoin");
					StringBuilder sb = new StringBuilder();
					p = globals[var[0]].Length - 1;
					for (int i = 0; i < p; i++)
						sb.Append(globals[var[0]][i]).Append(s);
					sb.Append(globals[var[0]][p]);
					r = sb.ToString();
					break;
				case _Token.STRLENGTH:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrLength");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "StrLength");
					r = r.Length.ToString(cul);
					break;
				case _Token.STRREMOVE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrRemove");
					r = expression();
					checktoken(_Token.COMMA, "StrRemove");
					s = expression();
					checktoken(_Token.COMMA, "StrRemove");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "StrRemove");
					r = r.Remove((int)double.Parse(s, cul), (int)double.Parse(t, cul)).ToString(cul);
					break;
				case _Token.STRREPLACE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrReplace");
					r = expression();
					checktoken(_Token.COMMA, "StrReplace");
					s = expression();
					checktoken(_Token.COMMA, "StrReplace");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "StrReplace");
					r = r.Replace(s, t);
					break;
				case _Token.STRTOLOWER:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrToLower");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "StrToLower");
					r = r.ToLower(cul);
					break;
				case _Token.STRTOUPPER:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "StrToUpper");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "StrToUpper");
					r = r.ToUpper(cul);
					break;
				case _Token.SUBSTRING:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "SubString");
					r = expression();
					checktoken(_Token.COMMA, "SubString");
					s = expression();
					checktoken(_Token.COMMA, "SubString");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "SubString");
					r = r.Substring((int)double.Parse(s, cul), (int)double.Parse(t, cul));
					break;
					// date functions
				case _Token.DATE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Date");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Date");
					r = new DateTime((long)double.Parse(r, cul)).ToString(dateformat, cul);
					break;
				case _Token.DATEADD:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateAdd");
					r = expression();
					checktoken(_Token.COMMA, "DateAdd");
					s = expression();
					checktoken(_Token.COMMA, "DateAdd");
					t = expression();
					checktoken(_Token.COMMA, "DateAdd");
					u = expression();
					checktoken(_Token.RIGHTPAREN, "DateAdd");
					DateTime date = new DateTime((long)double.Parse(r, cul));
					date = date.AddYears((int)double.Parse(s, cul)).AddMonths((int)double.Parse(t, cul)).AddDays(double.Parse(u, cul));
					r = date.Ticks.ToString(cul);
					break;
				case _Token.DATEDAY:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateDay");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateDay");
					DateTime time = new DateTime((long)double.Parse(r, cul));
					r = time.Day.ToString();
					break;
				case _Token.DATEDAYOFWEEK:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateDayOfWeek");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateDayOfWeek");
					time = new DateTime((long)double.Parse(r, cul));
					r = time.DayOfWeek.ToString();
					break;
				case _Token.DATEDAYOFYEAR:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateDayOfYear");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateDayOfYear");
					time = new DateTime((long)double.Parse(r, cul));
					r = time.DayOfYear.ToString(cul);
					break;
				case _Token.DATEMONTH:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateMonth");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateMonth");
					date = new DateTime((long)double.Parse(r, cul));
					r = date.Month.ToString(cul);
					break;
				case _Token.DATEPARSE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateParse");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateParse");
					{ try { r = DateTime.ParseExact(r, dateformat, cul).Ticks.ToString(cul); }
					  catch  { r = "0"; } }
					break;
				case _Token.DATEYEAR:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "DateYear");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "DateYear");
					date = new DateTime((long)double.Parse(r, cul));
					r = date.Year.ToString(cul);
					break;
					// time functions
				case _Token.TIME:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "Time");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "Time");
					r = new DateTime((long)double.Parse(r, cul)).ToString(timeformat, cul);
					break;
				case _Token.TIMEADD:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "TimeAdd");
					r = expression();
					checktoken(_Token.COMMA, "TimeAdd");
					s = expression();
					checktoken(_Token.COMMA, "TimeAdd");
					t = expression();
					checktoken(_Token.COMMA, "TimeAdd");
					u = expression();
					checktoken(_Token.RIGHTPAREN, "TimeAdd");
					time = new DateTime((long)double.Parse(r, cul));
					time = time.AddHours((int)double.Parse(s, cul)).AddMinutes((int)double.Parse(t, cul)).AddSeconds(double.Parse(u, cul));
					r = time.Ticks.ToString(cul);
					break;
				case _Token.TIMEPARSE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "TimeParse");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "TimeParse");
					{ try { r = DateTime.ParseExact(r, timeformat, cul).Ticks.ToString(cul);}
					  catch { r = "0"; } }
					break;
				case _Token.TIMEHOUR:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "TimeHour");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "TimeHour");
					time = new DateTime((long)double.Parse(r, cul));
					r = time.Hour.ToString(cul);
					break;
				case _Token.TIMEMINUTE:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "TimeMinute");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "TimeMinute");
					time = new DateTime((long)double.Parse(r, cul));
					r = time.Minute.ToString(cul);
					break;
				case _Token.TIMESECOND:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "TimeSecond");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "TimeSecond");
					time = new DateTime((long)double.Parse(r, cul));
					r = time.Second.ToString(cul);
					break;
					// calls
				case _Token.CALL:
					callstatement(false); //sets up for first line of sub
					do
					{
						if (ended)
							return "";
					} while (executeline(true) != _Token.RETURN);
					r = returnval;
					break;
				case _Token.CALLSUB:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "CallSub");
					subname = expression().ToLower(); // sub name
					callstatement(true); //sets up for first line of sub
					do
					{
						if (ended)
							return "";
					} while (executeline(true) != _Token.RETURN);
					r = returnval;
					break;
				case _Token.HOSTSUBCALL:
					u = tokeniserliteral(); // sub name
					getnexttoken();
					p = 0;
					r = s = t = "";
					if (currenttoken == _Token.LEFTPAREN)
					{
						{ p += 1; getnexttoken(); r = expression(); }
						if (currenttoken == _Token.COMMA)
						{ p += 1; getnexttoken(); s = expression(); }
						if (currenttoken == _Token.COMMA)
						{ p += 1; getnexttoken(); t = expression(); }
						if (currenttoken != _Token.RIGHTPAREN)
							saveerror(u + " : maximum of three parameters exceeded.");
						getnexttoken();
					}
					switch (p)
					{
						case 0: r = CallHostSub0(u, false); break;
						case 1: r = CallHostSub1(u, r, false); break;
						case 2: r = CallHostSub2(u, r, s, false); break;
						case 3: r = CallHostSub3(u, r, s, t, false); break;
					}
					break;
				case _Token.CALLHOSTSUB:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "CallHostSub");
					u = expression(); // sub name
					p = 0;
					r = s = t = "";
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); r = expression(); }
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); s = expression(); }
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); t = expression(); }
					if (currenttoken != _Token.RIGHTPAREN)
						saveerror("CallHostSub : maximum of three parameters exceeded.");
					getnexttoken();
					switch (p)
					{
						case 0: r = CallHostSub0(u, false); break;
						case 1: r = CallHostSub1(u, r, false); break;
						case 2: r = CallHostSub2(u, r, s, false); break;
						case 3: r = CallHostSub3(u, r, s, t, false); break;
					}
					break;
				case _Token.CALLHOSTGUISUB:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "CallHostGuiSub");
					u = expression(); // sub name
					p = 0;
					r = s = t = "";
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); r = expression(); }
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); s = expression(); }
					if (currenttoken == _Token.COMMA)
					{ p += 1; getnexttoken(); t = expression(); }
					if (currenttoken != _Token.RIGHTPAREN)
						saveerror("CallHostGuiSub : maximum of three parameters exceeded.");
					getnexttoken();
					switch (p)
					{
						case 0: r = CallHostSub0(u, true); break;
						case 1: r = CallHostSub1(u, r, true); break;
						case 2: r = CallHostSub2(u, r, s, true); break;
						case 3: r = CallHostSub3(u, r, s, t, true); break;
					}
					break;
				case _Token.RGB:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "RGB");
					r = s = t = "0";
					r = expression();
					checktoken(_Token.COMMA, "RGB");
					s = expression();
					checktoken(_Token.COMMA, "RGB");
					t = expression();
					checktoken(_Token.RIGHTPAREN, "RGB");
					r = Color.FromArgb((int)double.Parse(r), (int)double.Parse(s), (int)double.Parse(t)).ToArgb().ToString(cul);
					break;
				case _Token.GETB:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetB");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "GetB");
					color = Color.FromArgb((int)double.Parse(r));
					r = color.B.ToString();
					break;
				case _Token.GETG:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetG");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "GetG");
					color = Color.FromArgb((int)double.Parse(r));
					r = color.G.ToString();
					break;
				case _Token.GETR:
					getnexttoken();
					checktoken(_Token.LEFTPAREN, "GetR");
					r = expression();
					checktoken(_Token.RIGHTPAREN, "GetR");
					color = Color.FromArgb((int)double.Parse(r));
					r = color.R.ToString();
					break;
				case _Token.CBEIGE:
					r = "-526370";
					getnexttoken();
					break;
				case _Token.CBLACK:
					r = Color.Black.ToArgb().ToString(cul);
					getnexttoken();
					break;
				case _Token.CBLUE:
					r = Color.Blue.ToArgb().ToString(cul);
					getnexttoken();
					break;
				case _Token.CBROWN:
					r = "-5953495";
					getnexttoken();
					break;
				case  _Token.CCYAN:
					r = "-16711681";
					getnexttoken();
					break;
				case _Token.CGOLD:
					r = Color.Gold.ToArgb().ToString(cul);
					getnexttoken();
					break;
				case _Token.CGRAY:
					r = "-8093052";
					getnexttoken();
					break;
				case _Token.CGREEN:
					r = "-16743936";
					getnexttoken();
					break;
				case _Token.CORANGE:
					r = "-23040";
					getnexttoken();
					break;
				case _Token.CPINK:
					r = "-15410";
					getnexttoken();
					break;
				case _Token.CPURPLE:
					r = "-8126332";
					getnexttoken();
					break;
				case _Token.CRED:
					r = "-65536";
					getnexttoken();
					break;
				case _Token.CSILVER:
					r = "-3750970";
					getnexttoken();
					break;
				case _Token.CWHITE:
					r = Color.White.ToArgb().ToString(cul);
					getnexttoken();
					break;
				case _Token.CYELLOW:
					r = Color.Yellow.ToArgb().ToString(cul);
					getnexttoken();
					break;
				default:
					r = getvariable(tokeniservariable());
					checktoken(_Token.VARIABLE, "Variable");
				   break;
			}
			return r;
		}

		private string power()
		{
			string f1, f2;
			_Token op;
			f1 = factor();
			op = currenttoken;
			while (op == _Token.POWER)
			{
				getnexttoken();
				f2 = factor();
				f1 = (Math.Pow(double.Parse(f1, cul), double.Parse(f2, cul))).ToString(cul);
				op = currenttoken;
			}
			return f1;
		}

		private string multiplydivide()
		{
			string f1, f2;
			_Token op;
			f1 = power();
			op = currenttoken;
			while (op == _Token.TIMES ||
			  op == _Token.DIVIDE ||
			  op == _Token.MOD)
			{
				getnexttoken();
				f2 = power();
				switch (op)
				{
					case _Token.TIMES:
						f1 = (double.Parse(f1, cul) * double.Parse(f2, cul)).ToString(cul);
						break;
					case _Token.DIVIDE:
						f1 = (double.Parse(f1, cul) / double.Parse(f2, cul)).ToString(cul);
						break;
					case _Token.MOD:
						f1 = (double.Parse(f1, cul) % double.Parse(f2, cul)).ToString(cul);
						break;
				}
				op = currenttoken;
			}
			return f1;
		}

		private string addsubtract()
		{
			string t1, t2;
			_Token op;
			t1 = multiplydivide();
			op = currenttoken;
			while (op == _Token.PLUS ||
			  op == _Token.APPEND ||
			  op == _Token.MINUS ||
			  op == _Token.BITAND ||
			  op == _Token.BITOR ||
			  op == _Token.BITXOR)
			{
				getnexttoken();
				t2 = multiplydivide();
				switch (op)
				{
					case _Token.APPEND:
						t1 = t1 + t2;
						break;
					case _Token.PLUS:
						t1 = (double.Parse(t1, cul) + double.Parse(t2, cul)).ToString(cul);
						break;
					case _Token.MINUS:
						t1 = (double.Parse(t1, cul) - double.Parse(t2, cul)).ToString(cul);
						break;
					case _Token.BITAND:
						t1 = (((int)double.Parse(t1, cul)) & ((int)double.Parse(t2, cul))).ToString(cul);
						break;
					case _Token.BITOR:
						t1 = (((int)double.Parse(t1, cul)) | ((int)double.Parse(t2, cul))).ToString(cul);
						break;
					case _Token.BITXOR:
						t1 = (((int)double.Parse(t1, cul)) ^ ((int)double.Parse(t2, cul))).ToString(cul);
						break;
				}
				op = currenttoken;
			}
			return t1;
		}

		private string comparison()
		{
			string r1, r2;
			_Token op;
			r1 = addsubtract();
			op = currenttoken;
			while (op == _Token.LT ||
			  op == _Token.GT ||
			  op == _Token.EQ ||
			  op == _Token.NE ||
			  op == _Token.LE ||
			  op == _Token.GE)
			{
				getnexttoken();
				r2 = addsubtract();
				switch (op)
				{
					case _Token.LT:
						r1 = (double.Parse(r1, cul) < double.Parse(r2, cul) ? "true" : "false");
						break;
					case _Token.GT:
						r1 = (double.Parse(r1, cul) > double.Parse(r2, cul) ? "true" : "false");
						break;
					case _Token.EQ:
						r1 = (r1 == r2 ? "true" : "false");
						break;
					case _Token.NE:
						r1 = (r1 != r2 ? "true" : "false");
						break;
					case _Token.GE:
						r1 = (double.Parse(r1, cul) >= double.Parse(r2, cul) ? "true" : "false");
						break;
					case _Token.LE:
						r1 = (double.Parse(r1, cul) <= double.Parse(r2, cul) ? "true" : "false");
						break;
				}
				op = currenttoken;
			}
			return r1;
		}

		private string expression()
		{
			string r1, r2;
			_Token op;
			
			r1 = comparison();
			op = currenttoken;
			while (op == _Token.AND ||
			  op == _Token.OR)
			{
				getnexttoken();
				r2 = comparison();
				switch (op)
				{
					case _Token.AND:
						r1 = ((r1 == "true" && r2 == "true" ? "true" : "false"));
						break;
					case _Token.OR:
						r1 = ((r1 == "true" || r2 == "true" ? "true" : "false"));
						break;
				}
				op = currenttoken;
			}
			return r1;
		}
		
		// statements

		// when a statement is called currenttoken is the STATEMENT token except LET
		// when a statement returns currenttoken should be NUMBER for next line

		private _Token statement(bool infactor)
		{
			_Token token = currenttoken;
			switch (currenttoken)
			{
			case _Token.VARIABLE:
				// let is entered with currenttoken = VARIABLE
				letstatement();
				break;
			case _Token.LET:
				checktoken(_Token.LET, "Let"); // will succeed
				// let is entered with currenttoken = VARIABLE
				letstatement();
				break;
			case _Token.DIM:
				dimstatement();
				break;
			case _Token.DATEFORMAT:
				dateformatstatement();
				break;
			case _Token.TIMEFORMAT:
				timeformatstatement();
				break;
			case _Token.PRINT:
				printstatement();
				break;
			case _Token.MSGBOX:
				msgboxstatement();
				break;
			case _Token.CALL:
				callstatement(false);
				break;
			case _Token.CALLSUB:
				callsubstatement();
				break;
			case _Token.CALLHOST:
				callhoststatement();
				break;
			case _Token.HOSTSUBCALL:
				hostsubcallstatement();
				break;
			case _Token.CALLHOSTSUB:
				callhostsubstatement();
				break;
			case _Token.CALLHOSTGUISUB:
				callhostguisubstatement();
				break;
			case _Token.SUB:
				substatement();
				break;
			case _Token.GOTO:
				gotostatement();
				break;
			case _Token.RETURN:
				returnstatement(); // currentpointer at token after RETURN
				if (!infactor) // factor wants next token, execute line wants NUMBER
					getnexttoken();  // skip CR
				break;
			case _Token.FOR:
				forstatement();
				break;
			case _Token.NEXT:
				nextstatement();
				break;
			case _Token.DO:
				dostatement();
				break;
			case _Token.LOOP:
				loopstatement();
				break;
			case _Token.EXIT:
				exitstatement();
				break;
			case _Token.IF:
				ifstatement();
				break;
			case _Token.ELSE:
				elsestatement();
				break;
			case _Token.ELSEIF:
				elseifstatement();
				break;
			case _Token.ENDIF:
				endifstatement(); 
				break;
			case _Token.ENDSELECT:
				endselectstatement();               
				break;
			case _Token.ENDSUB:
				token = _Token.RETURN;
				endsubstatement();  // currentpointer at token after ENDSUB           
				if (!infactor) // factor wants next token, executeline wants NUMBER
					getnexttoken();  // skip CR
				break;
			case _Token.SELECT:
				selectstatement();
				break;
			case _Token.CASE:
				casestatement();
				break;
			case _Token.REM:
				remstatement();
				break;
			case _Token.APPCLOSE:
				appclosestatement();
				break;
			case _Token.SLEEP:
				sleepstatement();
				break;
			case _Token.ARRAY:
				arraystatement();
				break;
			case _Token.ARRAYCOPY:
				arraycopystatement();
				break;
			case _Token.STRSPLIT:
				strsplitstatement();
				break;
			case _Token.SETITEM:
				setitemstatement();
				break;
			case _Token.CR:
				getnexttoken(); // get NUMBER for next line
				break;
			default:
				saveerror("Statement : " + currenttoken + " not implemented ");
				break;
			}
			return token;
		}

		private void appclosestatement()
		{
			getnexttoken(); // CR
			checktoken(_Token.CR, "AppClose"); // at next line NUMBER token
			ended = true;
		}

		private void arraystatement()
		{
			int[] var;
			ArrayList list = new ArrayList();
			getnexttoken(); // skip ARRAY token
			checktoken(_Token.LEFTPAREN, "Array"); // skip LEFTPAREN token
			var = tokeniservariable();
			checktoken(_Token.VARIABLE, "Array"); // skip VARIABLE token
			do
			{
				getnexttoken(); // skip COMMA
				list.Add(expression());
			} while (currenttoken == _Token.COMMA);
			globals[var[0]] = new string[list.Count];
			globalsrank[var[0]] = new int[] { list.Count, 0, 0 };
			for (int i = 0; i < list.Count; i++)
				globals[var[0]][i] = (String) list[i];
			checktoken(_Token.RIGHTPAREN, "Array"); // skip RIGHTPAREN token
			checktoken(_Token.CR, "Array"); // at next line NUMBER token
		}

		private void arraycopystatement()
		{
			int[] src, target;
			int srcstart, len, targetstart;
			getnexttoken(); // skip ARRAYCOPY token
			checktoken(_Token.LEFTPAREN, "ArrayCopy"); // skip LEFTPAREN token
			src = tokeniservariable();
			checktoken(_Token.VARIABLE, "ArrayCopy"); // skip VARIABLE token
			checktoken(_Token.COMMA, "ArrayCopy");
			srcstart = (int)double.Parse(expression(), cul);
			checktoken(_Token.COMMA, "ArrayCopy");
			len = (int)double.Parse(expression(), cul);
			checktoken(_Token.COMMA, "ArrayCopy");
			target = tokeniservariable();
			checktoken(_Token.VARIABLE, "ArrayCopy"); // skip VARIABLE token
			checktoken(_Token.COMMA, "ArrayCopy");
			targetstart = (int)double.Parse(expression(), cul);
			for (int i = 0; i < len; i++)
				globals[target[0]][targetstart + i] = globals[src[0]][srcstart + i];
			checktoken(_Token.RIGHTPAREN, "ArrayCopy"); // skip RIGHTPAREN token
			checktoken(_Token.CR, "ArrayCopy"); // at next line NUMBER token
		}

		private void callstatement(bool callsub)
		{
			// CALL - next token is LEFTPAREN or something else
			// CALLSUB - next token is COMMA or RIGHTPAREN
			string calledsub;
			if (callsub)
				calledsub = subname;
			else
				calledsub = tokeniserliteral();
			// we need to increment callstackptr before evaluating expression() for parameters
			// because we might be re-entered if a sub call is used in a parameter
			
			if (subptrs.ContainsKey(calledsub))
			{
				if (callstackptr < MaxStackDepth)
				{
					callstackptr++;
					localsptr++;
				}
				else
				{
					saveerror("Call : stack depth exceeded, increase the StackDepth property");
					return;
				}
			}
			else
			{
				saveerror("Call : Sub " + calledsub + " not found");
				return;
			}
			parameters[callstackptr-1].Clear();
			if (!callsub)
				getnexttoken();
			if (currenttoken == _Token.LEFTPAREN ||
			   (callsub && (currenttoken == _Token.COMMA)))
			{
				do
				{
					getnexttoken(); // skip LEFTPAREN or COMMA
					localsptr--;
					parameters[callstackptr - 1].Add(expression());
					localsptr++;

				} while (currenttoken == _Token.COMMA);
				checktoken(_Token.RIGHTPAREN, "Sub parameters");
			}
			else
			{
				if (callsub)
					checktoken(_Token.RIGHTPAREN, "CallSub");
			}
			localsstack[callstackptr - 1] = localvars;
			callstack[callstackptr - 1] = tokenptr;
			gotolinenum((int)subptrs[calledsub]); // at next line NUMBER token
		}

		private void callsubstatement()
		{            
			getnexttoken();
			checktoken(_Token.LEFTPAREN, "CallSub");
			subname = expression().ToLower(); // sub name
			callstatement(true); 
		}

		private void callhoststatement()
		{
			getnexttoken(); // skip CALLHOST token
			checktoken(_Token.LEFTPAREN, "CallHost"); // skip LEFTPAREN token
			messagestring = expression();
			subjectstring = "";
			if (currenttoken == _Token.COMMA)
			{
				getnexttoken(); // skip COMMA token
				subjectstring = expression();
			}
			CallHostEvent();
			checktoken(_Token.RIGHTPAREN, "CallHost"); // skip RIGHTPAREN token
			checktoken(_Token.CR, "CallHost"); // at next line NUMBER token
		}

		private void callhostsubstatement()
		{
			callhostsubbasestatement(false, "CallHostSub");
		}

		private void callhostguisubstatement()
		{
			callhostsubbasestatement(true, "CallHostGuiSub");
		}

		private void callhostsubbasestatement(bool invoke, string name)
		{
			string u, r, s, t;
			getnexttoken(); // skip CALLHOSTSUB token
			checktoken(_Token.LEFTPAREN, name);
			u = expression(); // sub name
			int p = 0;
			r = s = t = "";
			if (currenttoken == _Token.COMMA)
			{
				p += 1;
				getnexttoken();
				r = expression();
			}
			if (currenttoken == _Token.COMMA)
			{
				p += 1;
				getnexttoken();
				s = expression();
			}

			if (currenttoken == _Token.COMMA)
			{
				p += 1;
				getnexttoken();
				t = expression();
			}
			if (currenttoken != _Token.RIGHTPAREN)
				saveerror(name + " : maximum of three parameters exceeded.");
			getnexttoken();
			switch (p)
			{
				case 0:
					r = CallHostSub0(u, invoke);
					break;
				case 1:
					r = CallHostSub1(u, r, invoke);
					break;
				case 2:
					r = CallHostSub2(u, r, s, invoke);
					break;
				case 3:
					r = CallHostSub3(u, r, s, t, invoke);
					break;
			}
			checktoken(_Token.CR, name); // at next line NUMBER token
		}

		private void casestatement()
		{
			int caseline = (int)source[++tokenptr];
			getnexttoken(); // skip CASE token
			if (matchedcases[callstackptr])
				gotolinenum(endselectlines[callstackptr]); // jump if already matched
			else if (currenttoken != _Token.ELSE)
			{
				String testval = expression(); // save value
				checktoken(_Token.CR, "Case");
				if (selectvals[callstackptr] != testval)
				{
					if (caseline != 0)
						gotolinenum(caseline);
					else
						gotolinenum(endselectlines[callstackptr]);
				}
				else
				{
					matchedcases[callstackptr] = true;
				}
			}
			else
			{
				getnexttoken(); // skip ELSE
				checktoken(_Token.CR, "Case");
			}
		}

		private void dateformatstatement()
		{
			getnexttoken(); // skip DATEFORMAT token
			checktoken(_Token.LEFTPAREN, "Date");
			dateformat = expression();
			checktoken(_Token.RIGHTPAREN, "Date");
			checktoken(_Token.CR, "Date"); // at next line NUMBER token
		}

		private void dimstatement()
		{
			int[] var;
			getnexttoken(); // skip DIM token
			if (currenttoken == _Token.TYPE)
			{
				typestatement();
				return;
			}
			var = tokeniservariable();
			checktoken(_Token.VARIABLE, "Dim variable");
			if (var[1] >= 0)
				saveerror("Dim : Cannot Dim a local variable");
			// [2] is rightmost, [3] is middle, [4] is leftmost dimension
			globalsrank[var[0]] = new int[] { var[2], var[3], var[4] };
			var[2] = var[2] == 0 ? 1 : var[2];
			var[3] = var[3] == 0 ? 1 : var[3];
			var[4] = var[4] == 0 ? 1 : var[4];
			globals[var[0]] = new string[var[2] * var[3] * var[4]];
			checktoken(_Token.CR, "Dim"); // at next line NUMBER token
		}


		private void typestatement()
		{
			int[] var;
			int index = 0;
			getnexttoken(); // skip TYPE token
			checktoken(_Token.LEFTPAREN, "Type"); // skip LEFTPAREN token
			do
			{
				if (index > 0)
					getnexttoken(); // skip COMMA
				var = tokeniservariable();
				setvariable(var, index.ToString(cul)); // set indexer variable
				index++;
				getnexttoken(); // skip Variable
			} while (currenttoken == _Token.COMMA);
			checktoken(_Token.RIGHTPAREN, "Type"); // skip RIGHTPAREN token
			var = tokeniservariable();
			checktoken(_Token.VARIABLE, "Dim Type variable");
			if (var[1] >= 0)
				saveerror("Dim : Cannot Dim a local variable");
			// [2] is rightmost, [3] is middle, [4] is leftmost dimension
			globalsrank[var[0]] = new int[] { index, var[2], var[3] };
			var[2] = var[2] == 0 ? 1 : var[2];
			var[3] = var[3] == 0 ? 1 : var[3];
			globals[var[0]] = new string[index * var[2] * var[3]];
			checktoken(_Token.CR, "Dim Type"); // at next line NUMBER token

		}

		private void dostatement()
		{
			int exitline = (int)source[++tokenptr];
			getnexttoken();
			if (currenttoken == _Token.CR)
			{
				// must be LOOP WHILE/UNTIL
				checktoken(_Token.CR, "Do"); // at next line NUMBER token 
				if (forstackptr[callstackptr] < MaxStackDepth)
				{
					forstack[callstackptr, forstackptr[callstackptr]].exitline = exitline;
					forstack[callstackptr, forstackptr[callstackptr]].loopline = (int)source[tokenptr]; // next line number
					forstack[callstackptr, forstackptr[callstackptr]].forvariable = new int[2] { 0, 0 };
					ifstackptr[callstackptr] = forstack[callstackptr, forstackptr[callstackptr]].ifstackptr; // saved in case Exit in If block
					forstack[callstackptr, forstackptr[callstackptr]].to = 0;
					forstack[callstackptr, forstackptr[callstackptr]].step = 0;
					forstackptr[callstackptr]++;
				}
				else
				{
					saveerror("Do : stack depth exceeded, increase the StackDepth property");
				}
			}
			else
			{                
				_Token looptoken = currenttoken; // should be WHIILE or LOOP
				if (looptoken != _Token.WHILE && looptoken != _Token.UNTIL)
					saveerror("Do : unrecognized token " + looptoken);
				else
				{
					getnexttoken();
					int dolinenum = linenum; // save linenum
					string val = expression(); // could alter linenum if Sub call
					checktoken(_Token.CR, "Do"); // at next line NUMBER token 
					if ((val == "true" && looptoken == _Token.UNTIL) ||
						(val == "false" && looptoken == _Token.WHILE))
					{
						gotolinenum(exitline); // at next line NUMBER token
					}
					else
					{
						if (forstackptr[callstackptr] < MaxStackDepth)
						{
							forstack[callstackptr, forstackptr[callstackptr]].exitline = exitline;
							forstack[callstackptr, forstackptr[callstackptr]].loopline = dolinenum;
							forstack[callstackptr, forstackptr[callstackptr]].forvariable = new int[2] { 0, 0 };
							ifstackptr[callstackptr] = forstack[callstackptr, forstackptr[callstackptr]].ifstackptr; // saved in case Exit in If block
							forstack[callstackptr, forstackptr[callstackptr]].to = 0;
							forstack[callstackptr, forstackptr[callstackptr]].step = 0;
							forstackptr[callstackptr]++;
						}
						else
						{
							saveerror("Do : stack depth exceeded, increase the StackDepth property");
						}
					}
				}
			}
		}

		private void elsestatement()
		{
			if (ifstack[callstackptr, ifstackptr[callstackptr] - 1].val == "true")
			{
				gotolinenum(ifstack[callstackptr, ifstackptr[callstackptr] - 1].endifline);
			}
			else
			{
				getnexttoken();  //skip ELSE token
				checktoken(_Token.CR, "Else"); // at next line NUMBER token
			}
		}

		private void elseifstatement()
		{
			String val;
			int elseline = (int)source[++tokenptr];
			if (ifstack[callstackptr, ifstackptr[callstackptr] - 1].val == "true")
			{
				gotolinenum(ifstack[callstackptr, ifstackptr[callstackptr] - 1].endifline);
				return;
			}
			getnexttoken();  //skip ELSEIF token
			val = expression();
			checktoken(_Token.THEN, "Else If");
			checktoken(_Token.CR, "Else If");
			ifstack[callstackptr, ifstackptr[callstackptr] - 1].val = val;
			if (val != "true")
			{
				if (elseline > 0)
					gotolinenum(elseline);
				else
				{
					 gotolinenum(ifstack[callstackptr, ifstackptr[callstackptr] - 1].endifline);
				}
			}
		}

		private void endifstatement()
		{
			ifstackptr[callstackptr]--;
			getnexttoken();
			checktoken(_Token.CR, " End If"); // at next line NUMBER token
		}

		private void endselectstatement()
		{
			getnexttoken();
			checktoken(_Token.CR, " End Select"); // at next line NUMBER token
		}

		private void endsubstatement()
		{
			returnval = "";
			getnexttoken();
			checktoken(_Token.CR, "End Sub"); // skip CR, at NUMBER
			if (callstackptr > 0)
			{
				ifstackptr[callstackptr] = 0;  // reset stackptrs, maybe not necessary here at End Sub
				forstackptr[callstackptr] = 0; // in case we return in a loop or if
				callstackptr--;
				localsptr--;
				localvars = localsstack[callstackptr];
				gotoplace(callstack[callstackptr]);
			}
			else
			{
				saveerror("End Sub : stack empty");
			}
		}

		private void exitstatement()
		{
			getnexttoken();
			if (forstackptr[callstackptr] > 0)
			{
				forstackptr[callstackptr]--;
				int exitline = forstack[callstackptr, forstackptr[callstackptr]].exitline;
				ifstackptr[callstackptr] = forstack[callstackptr, forstackptr[callstackptr]].ifstackptr; // saved in case Exit in If block
				gotolinenum(exitline); // at next line NUMBER token
			}
			else
			{
				saveerror("Exit : stack empty");
				checktoken(_Token.CR, "Exit"); // at next line NUMBER token
			}
		}

		private void forstatement()
		{
			int[] forvariable;
			double to, step = 1; // default step value
			int exitline = (int)source[++tokenptr];
			getnexttoken();
			forvariable = tokeniservariable();
			checktoken(_Token.VARIABLE, "For");
			checktoken(_Token.EQ, "For");
			setvariable(forvariable, expression());
			checktoken(_Token.TO, "To");
			to = double.Parse(expression(), cul);
			if (currenttoken == _Token.STEP)
			{
				getnexttoken();                
				step = double.Parse(expression(), cul);
			}
			checktoken(_Token.CR, "For"); // at next line NUMBER token 
			if(forstackptr[callstackptr] < MaxStackDepth)
			{
				double startval = double.Parse(getvariable(forvariable), cul);
				if ((step > 0 && startval > to) || (step < 0 && startval < to))
				{
					gotolinenum(exitline);
					return;
				}                
				int saveptr = tokenptr;
				forstack[callstackptr, forstackptr[callstackptr]].exitline = exitline;
				forstack[callstackptr, forstackptr[callstackptr]].loopline = (int)source[tokenptr];
				forstack[callstackptr, forstackptr[callstackptr]].ifstackptr = ifstackptr[callstackptr]; // saved in case Exit within If block
				tokenptr = saveptr;
				forstack[callstackptr, forstackptr[callstackptr]].forvariable = forvariable;
				forstack[callstackptr, forstackptr[callstackptr]].to = to;
				forstack[callstackptr, forstackptr[callstackptr]].step = step;
				forstackptr[callstackptr]++;
			}
			else
			{
				saveerror("For : stack depth exceeded, increase the StackDepth property");
			}
		}

		private void gotostatement()
		{
			forstackptr[callstackptr] += (Int16)source[++tokenptr];
			ifstackptr[callstackptr] += (Int16)source[++tokenptr];
			int strlen = source[++tokenptr];
			string label = source.Substring(++tokenptr, strlen);
			object labellinenum = labelptrs[label];
			gotolinenum((int)labellinenum);
		}
		
		private void hostsubcallstatement()
		{
			String calledsub;
			calledsub = tokeniserliteral();
			string r, s, t;
			getnexttoken(); // skip CALLHOSTSUB token
			int p = 0;
			r = s = t = "";
			if (currenttoken == _Token.LEFTPAREN)
			{
				{
					p += 1;
					getnexttoken();
					r = expression();
				}
				if (currenttoken == _Token.COMMA)
				{
					p += 1;
					getnexttoken();
					s = expression();
				}

				if (currenttoken == _Token.COMMA)
				{
					p += 1;
					getnexttoken();
					t = expression();
				}
				if (currenttoken != _Token.RIGHTPAREN)
					saveerror(calledsub + " : maximum of three parameters exceeded.");
				getnexttoken();
			}
			switch (p)
			{
				case 0:
					r = CallHostSub0(calledsub, false);
					break;
				case 1:
					r = CallHostSub1(calledsub, r, false);
					break;
				case 2:
					r = CallHostSub2(calledsub, r, s, false);
					break;
				case 3:
					r = CallHostSub3(calledsub, r, s, t, false);
					break;
			}
			checktoken(_Token.CR, calledsub); // at next line NUMBER token
		}

		private void ifstatement()
		{
			string val;
			int elseline = (int)source[++tokenptr];
			int endifline = (int)source[++tokenptr];
			getnexttoken(); //skip IF token
			val = expression();
			checktoken(_Token.THEN, "Then");
			if (currenttoken != _Token.CR)
			{
				saveerror("Then : expected CR found " + currenttoken);
			}
			else
			{
				if (endifline == 0)
					saveerror("If : End If not found");
				if (ifstackptr[callstackptr] < MaxStackDepth)
				{
					ifstack[callstackptr, ifstackptr[callstackptr]].endifline = endifline;
					ifstack[callstackptr, ifstackptr[callstackptr]].val = val;
					ifstackptr[callstackptr]++;
				}
				else
				{
					saveerror("If : stack depth exceeded, increase the StackDepth property");
				}
				if (val != "true")
				{
					if (elseline > 0)
						gotolinenum(elseline);
					else
						gotolinenum(endifline);
				}
				else
					checktoken(_Token.CR, "If"); // at next line NUMBER token
			}
		}

		private void letstatement()
		{
			int[] var;
			var = tokeniservariable();
			checktoken(_Token.VARIABLE, "Assign variable");
			checktoken(_Token.EQ, "Assign");
			setvariable(var, expression());
			checktoken(_Token.CR, "Assign"); // at next line NUMBER token
		}

		private void loopstatement()
		{
			getnexttoken();
			_Token looptoken = currenttoken;
			if (looptoken == _Token.CR)
			{
				gotolinenum(forstack[callstackptr, forstackptr[callstackptr] - 1].loopline); // at next line NUMBER token
				forstackptr[callstackptr]--; // DO WHILE/UNTIL will put new frame on stack
				return;
			}
			if (looptoken == _Token.UNTIL ||
				looptoken == _Token.WHILE)
			{
				getnexttoken();
				string val = expression();
				if(forstackptr[callstackptr] > 0)
				{
					if ((val == "true" && looptoken == _Token.WHILE) ||
						(val == "false" && looptoken == _Token.UNTIL))
					{
						gotolinenum(forstack[callstackptr, forstackptr[callstackptr] - 1].loopline); // at next line NUMBER token
					}
					else 
					{
					  forstackptr[callstackptr]--;
					  checktoken(_Token.CR, "Loop"); // at next line number
				  }
				}
				else
				{
					saveerror("Loop : stack empty");
				}
			}
			else
			{
				saveerror("Loop : unexpected token " + currenttoken);
			}
		}        

		private void msgboxstatement()
		{
			getnexttoken(); // skip MSGBOX token
			checktoken(_Token.LEFTPAREN, "MsgBox"); // skip LEFTPAREN token
			messagestring = expression();
			subjectstring = "";
			if (currenttoken == _Token.COMMA)
			{
				getnexttoken(); // skip COMMA token
				subjectstring = expression();
			}
			MessageEvent();           
			checktoken(_Token.RIGHTPAREN, "MsgBox"); // skip RIGHTPAREN token
			checktoken(_Token.CR, "MsgBox"); // at next line NUMBER token
		}

		private void nextstatement()
		{
			double val;
			getnexttoken();
			if (forstackptr[callstackptr] > 0)
			{
				int[] var = forstack[callstackptr, forstackptr[callstackptr] - 1].forvariable;
				double to = forstack[callstackptr, forstackptr[callstackptr] - 1].to;
				double step = forstack[callstackptr, forstackptr[callstackptr] - 1].step;

				val = double.Parse(getvariable(var), cul) + step;
				setvariable(var, val.ToString(cul));


				if (step > 0 && val <= to )
				{
					gotolinenum(forstack[callstackptr, forstackptr[callstackptr] - 1].loopline); // at next line NUMBER token
				}
				else if (step < 0 && val >= to)
					{
						gotolinenum(forstack[callstackptr, forstackptr[callstackptr] - 1].loopline); // at next line NUMBER token
					}
					else
					{
						forstackptr[callstackptr]--;
						checktoken(_Token.CR, "Next");  // at next line NUMBER token
					}
			}
			else
			{
				saveerror("Next : stack empty");
				checktoken(_Token.CR, "Next"); // at next line NUMBER token
			}
	  }

		private void printstatement()
		{
			string str = "";
			getnexttoken(); // skip PRINT token
			do
			{
				switch (currenttoken)
			{
				case _Token.COMMA:
					str = str + "\t";
					getnexttoken();
					break;
				case _Token.SEMICOLON:
					getnexttoken();
					break;
				default:
					str = str + expression();
					break;
				}
			} while(currenttoken != _Token.CR && currenttoken != _Token.ENDOFINPUT);
			messagestring = str;
			subjectstring = "Print";
			MessageEvent();
			checktoken(_Token.CR, "Print"); // at next line NUMBER token
		}

		private void returnstatement()
		{
			getnexttoken();  // skip RETURN token
			if (currenttoken != _Token.CR)
				returnval = expression();
			else
				returnval = "";            
			checktoken(_Token.CR, "Return"); // skip CR, at NUMBER
			if (callstackptr > 0)
			{
				ifstackptr[callstackptr] = 0;  // reset stackptrs
				forstackptr[callstackptr] = 0; // in case we return in a loop or If
				callstackptr--;
				localsptr--;
				localvars = localsstack[callstackptr];
				gotoplace(callstack[callstackptr]); 
			}
			else
			{
				saveerror("Return : stack empty");
			}
		}

		private void remstatement()
		{
			getnexttoken(); // at CR
			checktoken(_Token.CR, "Rem"); 
		}

		private void selectstatement()
		{
			int caseline = (int)source[++tokenptr];
			int endselectline = (int)source[++tokenptr];
			getnexttoken(); // skip SELECT token
			selectvals[callstackptr] = expression(); // save value
			matchedcases[callstackptr] = false;
			if (currenttoken != _Token.CR)
			{
				saveerror("Select : expected CR found " + currenttoken);
			}
			else
			{
			   // getnexttoken(); // should be NUMBER
				endselectlines[callstackptr] = endselectline;
				if (endselectlines[callstackptr] == 0)
					saveerror("Select : End Select not found");
				else if (caseline == 0)
					saveerror("Select : Case not found");
				else
					gotolinenum(caseline);
			}
		}

		private void setitemstatement()
		{
			string r, s, t;
			getnexttoken(); // skip SETITEM token
			checktoken(_Token.LEFTPAREN, "SetItem");
			r = expression();
			checktoken(_Token.COMMA, "SetItem");
			s = expression();
			checktoken(_Token.COMMA, "SetItem");
			t = expression();
			globals[(int)globalvars[r]][(int)double.Parse(s, cul)] = t;
			checktoken(_Token.RIGHTPAREN, "SetItem");
			checktoken(_Token.CR, "SetItem"); // at next line NUMBER token
		}


		private void sleepstatement()
		{
			getnexttoken(); // skip SLEEP token
			checktoken(_Token.LEFTPAREN, "Sleep"); // skip LEFTPAREN token
			string s = expression();
			checktoken(_Token.RIGHTPAREN, "Sleep"); // skip RIGHTPAREN token
			System.Threading.Thread.Sleep((int)double.Parse(s, cul));
			checktoken(_Token.CR, "Sleep"); // at next line NUMBER token
		}

		private void strsplitstatement()
		{
			int[] var;
			String str, seps;
			String[] res;
			getnexttoken(); // skip STRSPLIT token
			checktoken(_Token.LEFTPAREN, "StrSplit"); // skip LEFTPAREN token
			var = tokeniservariable();
			checktoken(_Token.VARIABLE, "StrSplit"); // skip VARIABLE token
			checktoken(_Token.COMMA, "StrSplit");
			str = expression();
			checktoken(_Token.COMMA, "StrSplit");
			seps = expression();
			res = str.Split(seps.ToCharArray());
			globals[var[0]] = new string[res.Length];
			globalsrank[var[0]] = new int[] { res.Length, 0, 0 };
			for (int i = 0; i < res.Length; i++)
				globals[var[0]][i] = res[i];
			checktoken(_Token.RIGHTPAREN, "StrSplit"); // skip RIGHTPAREN token
			checktoken(_Token.CR, "StrSplit"); // at next line NUMBER token
		}

		private void substatement()
		{
			if (callstackptr == 0)
			{
				ended = true;
				return;
			}
			String name = tokeniserliteral();  // SUB name
			localvars = (Hashtable) subvars[name];
			getnexttoken();  // CR or LEFTPAREN
			if (currenttoken == _Token.LEFTPAREN)
			{
				int[] var;
				int i = 0;
				do
				{
					getnexttoken(); // VARIABLE
					var = tokeniservariable();
					checktoken(_Token.VARIABLE, "Sub"); //COMMA or RIGHTPAREN
					setvariable(var, (String)parameters[callstackptr - 1][i++]);
				} while (currenttoken == _Token.COMMA);
				getnexttoken();  // CR 
			}
			checktoken(_Token.CR, "Sub");
		}     

		private void timeformatstatement()
		{
			getnexttoken(); // skip TIMEFORMAT token
			checktoken(_Token.LEFTPAREN, "Time");
			timeformat = expression();
			checktoken(_Token.RIGHTPAREN, "Time");
			_Token lasttoken = currenttoken; // CR 
			checktoken(_Token.CR, "Time"); // at next line NUMBER token
		}      
				
#endregion Interpreter


#region Tokeniser

		//these routines deal with moving along the source code stream using
		//  currenttoken - the present token
		//  tokenptr  - pointer to the start of the present token
		//  nextptr  - pointer to the start of the present token
	   
		// gets first token in the source stream, sets current_token
		private void tokeniserinit()
		{
			tokenptr = nextptr = 0;
			fetchnextoken(); // using tokenptr not nextptr
		}

		// identifies next token at tokenptr and sets current_token
		// tokenptr at start of new token, nextptr at end+1
		private void fetchnextoken()
		{
			tokenptr = nextptr;
			while (source[tokenptr] == ' ' || source[tokenptr] == '\t')
			{
				++tokenptr;
			}
			int i;
			char chr = source[tokenptr];
			if(chr == 0)
			{
				currenttoken =  _Token.ENDOFINPUT;
				return ;
			}
			if (chr == '\r')
			{
				nextptr = tokenptr + 2;
				currenttoken =  _Token.CR;
				return;
			}
			// numeric literals are decimals with an optional decimal point
			// no longer allows a number to start with a decimal point because of array.access
			// scientific notation is acepted
			if (Char.IsDigit(chr)) 
			{
				i = 0;
				do
				{
					i++;
					chr = source[tokenptr + i];
					// skip "e" and exponent sign or first number
					if (Char.ToLower(chr) == 'e')
					{
						i += 2;
						chr = source[tokenptr + i];
					}

				} while (Char.IsDigit(chr) ||  chr == '.');
				nextptr = tokenptr + i;
				currenttoken = _Token.NUMBER;
				return;
			}

			// string is delimited by " or '
			if (chr == '"' || chr == '\'')
			{
				char ch;
				nextptr = tokenptr;
				do
				{
					++nextptr;
					ch = source[nextptr];
				//} while (ch != '"' && ch != '\'');
				} while (ch != chr);
				++nextptr;
				currenttoken =  _Token.STRING;
				return;
			}
			object keyobj;
			object subobj;
			object hostsubobj;
			// special case for non-numeric 2 character keywords
			// first two letters of a variable must not be seen as a keyword
			string op2 =  source.Substring(tokenptr, 2);
			keyobj = keywords[source.Substring(tokenptr, 2)];
			if (keyobj != null)
			{

				currenttoken = (_Token)keyobj;
				if (currenttoken == _Token.NE ||
					currenttoken == _Token.LE ||
					currenttoken == _Token.GE)
				{
					nextptr = tokenptr + 2;
					return;
				}
			}
			// special case for non-numeric 1 character keywords
			// all one letter keywords are non-(digit or letter)
			keyobj = keywords[source.Substring(tokenptr, 1)];
			if (keyobj != null)
			{
				nextptr = tokenptr + 1;               
				currenttoken =  (_Token) keyobj;
				return;
			}
			// all other keywords look like valid names
			int j = tokenptr;
			do
			{
				j++;
			} while (Char.IsLetterOrDigit(source[j])
				|| source[j] == '_'
				|| source[j] == '#');
			int length = j - tokenptr;
			String name = source.Substring(tokenptr, length).ToLower(cul);
			keyobj = keywords[name];
			subobj = subptrs[name];
			hostsubobj = hostsubnames[name];
			if (keyobj != null)
			{
				nextptr = tokenptr + length;
				currenttoken =  (_Token)keyobj;
				return;
			}
			if (hostsubobj != null)
			{
				subname = name;
				nextptr = tokenptr + length;
				currenttoken = (_Token) hostsubobj;
				return;
			}
			if (subobj != null)
			{
				subname = name;
				nextptr = tokenptr + length;
				currenttoken = _Token.CALL;
				return;
			}
			chr = Char.ToLower(chr);
			if ((chr >= 'a' && chr <= 'z') || chr == '_')
			{
				nextptr = tokenptr + 1;
				currenttoken = _Token.VARIABLE;
				return; 
			}
			saveerror("Tokeniser : unrecognized token beginning with " + chr + " - Chr(" + (int) chr + ")");
			currenttoken = _Token.ERROR;
			return;
		}

#endregion Tokeniser


#region Token dispatcher

	  //these routines deal with moving along the token stream using
	  //  currenttoken - the present token
	  //  tokenptr  - pointer to the start of the present token

	  // gets first token in the token stream, sets current_token
	  private void tokendispatcherinit()
	  {
		  tokenptr = -1;
		  getnexttoken(); 
	  }

	  // checks currenttoken, if no error gets nexttoken
	  private void checktoken(_Token token, string sourceid)
	  {
		  if (token != currenttoken)
		  {
			  saveerror(sourceid + " : expected " + token + ", found " + currenttoken);
		  }
		  currenttoken = (_Token)source[++tokenptr];
	  }

	  // identifies next token at tokenptr and sets current_token
	  // tokenptr at new token
	  private void getnexttoken()
	  {
		  currenttoken = (_Token) source[++tokenptr];
	  }

	  // returns a number from the token stream
	  private double tokenisernum()
	  {
		  return double.Parse(tokeniserliteral(), cul);
	  }

	  // returns a string from the token stream
	  private String tokeniserliteral()
	  {
		  int index = (int)source[++tokenptr];
		  String str = (string) literals[index];
		  return str;
	  }

	  private int[] tokeniservariable()
	  {
		// [0] is variable, [1] is local = 1, global = -1 
		// [2] is rightmost, [3] is middle, [4] is lefttmost dimension
		  int[] var = new int[] { 0, -1, 0, 0, 0 };
		if (source[++tokenptr] == 'L')
		{
			var[1] = 1; 
		}
		var[0] = (int)source[++tokenptr];
		if ((_Token)source[tokenptr + 1] == _Token.LEFTPAREN)
		{
			if (var[1] == 1)
				saveerror("Array : local variable cannot be an array");
			checktoken(_Token.VARIABLE, "Array");
			checktoken(_Token.LEFTPAREN, "Array");
			if (currenttoken != _Token.RIGHTPAREN)
			{
				var[2] = (int)double.Parse(expression(), cul);
				if (currenttoken == _Token.COMMA)
				{
					getnexttoken();
					var[3] = var[2];
					var[2] = (int)double.Parse(expression(), cul);
					if (currenttoken == _Token.COMMA)
					{
						var[4] = var[3];
						var[3] = var[2];
						getnexttoken();
						var[2] = (int)double.Parse(expression(), cul);
					}
				}
			}
			checktoken(_Token.RIGHTPAREN, "Array");
			tokenptr--;
			currenttoken = _Token.VARIABLE;
		}
		else if ((_Token)source[tokenptr + 1] == _Token.DOT)
		{
			if (var[1] == 1)
				saveerror("Array : local variable cannot be an array");
			checktoken(_Token.VARIABLE, "Array");
			checktoken(_Token.DOT, "Array");
			if (currenttoken == _Token.NUMBER)
			{
				var[4] = var[3];
				var[3] = var[2];
				var[2] = (int) tokenisernum();
				currenttoken = _Token.VARIABLE; // expected after call to tokeniservariable
				return var;
			}
			int[] var2 = tokeniservariable();
			var[4] = var[3];
			var[3] = var[2];
			var[2] = (int)double.Parse(getvariable(var2));
			currenttoken = _Token.VARIABLE;
		}
		return var;
	  }

#endregion Token dispatcher

  }
}